(********************************************)
(* SVGA.TPU vs. GRAPH.TPU: The Profiling    *)
(********************************************)
(* Example program for VRg SuperVGA library *)
(* Copyleft (C) 1996-1999 Virtual Research  *)
(********************************************)

program svgaprof;
{$S-,Q-,R-,V-,B-,X+,G+,N+}
uses Graph,SVGA,CRT,DOS;
var driver:integer;
    ticks:longint;
    logf:text;
    x1,x2,y1,y2,x3,y3:word;
    i:longint;
    c,m:byte;
    mode:integer;
    p1,p2:pointer;
    Year,Month,Day,DayOfWeek:Word;
    Machine,VCard:string;

const days:array[0..6] of string[3]=('Sun','Mon','Tue','Wed','Thr','Fri','Sat');
      mont:array[1..12] of string[3]=
   ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');

procedure xchange(var a,b:word);
var c:word;
begin
 c:=a;
 a:=b;
 b:=c;
end;

procedure TimerReset;
begin
 ticks:=MemL[$0040:$006C];
end;

function TimerRead:real;
begin
 TimerRead:=(MemL[$0040:$006C]-ticks)/18.2;
end;

procedure FONT;external;
{$L FONT.OBJ}

procedure VESA;external;
{$L VESA.OBJ}

begin
 clrscr;
 writeln (#10,#13,'SuperVGA graphics routines performance measurment tool');
 writeln('(C)oded by Asp / VR group, 1998-1999');
 writeln('This program will compare performance of VR group SuperVGA library');
 writeln('and BGI graphics (VESA.BGI by Jordan Hargrafix).',#10,#13);
 writeln('Enter your machine description, please:');
 readln(Machine);
 writeln('Enter your videocard description, please:');
 readln(VCard);
 writeln(#10,#13,'Tests will take about 5..20 minutes depending on your hardware.');
 writeln('Tests results will be written to SVGAPROF.LOG in current directory.');
 writeln('If there will be no screen activity for long time, please reboot your system.');
 writeln('Press a key to start tests now.',#10,#13);
 ReadKey;

 if not _VESAPresent then begin
  writeln('VESA BIOS extensions not found.');
  halt(1);
 end;
 assign(logf,'svgaprof.log');
 rewrite(logf);
 GetDate(Year,Month,Day,DayOfWeek);
 writeln(logf,'SVGAPROF log file');
 writeln(logf,'=================');
 writeln(logf,SVGA.Copyleft_Id);
 write(logf,'SVGAPROF log file generated on ',days[DayOfWeek],' ',Day,'.',mont[Month],'.',Year,', ');
 GetTime(Year,Month,Day,DayOfWeek);
 writeln(logf,Year,':',Month,':',Day);
 writeln(logf,'Machine: '+Machine,#13,#10,'Video: '+VCard,#13,#10);
 driver:=InstallUserDriver('VESA',nil);
 if RegisterBGIdriver(@VESA)<0 then begin
  writeln(#10,#13,'Cannot init graphics driver (EXE corrupted?)');
  halt(1);
 end;
 mode:=2;
 InitGraph(driver,mode,'');
 _InitMode(_640x480);
 randomize;

 (* Ellipse profiling *)
 writeln(logf,'Ellipse: 15000 calls');
 timerReset;
 for i:=1 to 15000 do begin
  SetColor(random(256));
  Ellipse(random(440)+100,random(280)+100,0,360,random(100),random(100));
 end;
 writeln(logf,'BGI Ellipse: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 15000 do
  _Ellipse(random(440)+100,random(280)+100,random(100),random(100),random(256));
 writeln(logf,'_Ellipse: ',timerRead:8:3,' sec');
 ClearDevice;

(* Rectangle profiling *)
 writeln(logf,'Rectangle: 15000 calls');
 timerReset;
 for i:=1 to 15000 do begin
  x1:=random(640);
  y1:=random(480);
  x2:=random(640);
  y2:=random(480);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  SetColor(random(256));
  Rectangle(x1,y1,x2,y2);
 end;
 writeln(logf,'BGI Rectangle: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 15000 do begin
  x1:=random(640);
  y1:=random(480);
  x2:=random(640);
  y2:=random(480);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  _Box(x1,y1,x2,y2,random(256),0);
 end;
 writeln(logf,'_Box: ',timerRead:8:3,' sec');
 ClearDevice;

(* Bar profiling *)
 writeln(logf,'Bar: 15000 calls');
 timerReset;
 for i:=1 to 15000 do begin
  x1:=random(640);
  y1:=random(480);
  x2:=random(640);
  y2:=random(480);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  SetFillStyle(1,random(256));
  Bar(x1,y1,x2,y2);
 end;
 writeln(logf,'BGI Bar: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 15000 do begin
  x1:=random(640);
  y1:=random(480);
  x2:=random(640);
  y2:=random(480);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  c:=random(255)+1;
  _Box(x1,y1,x2,y2,c,c);
 end;
 writeln(logf,'_Box: ',timerRead:8:3,' sec');
 ClearDevice;

(* PutPixel profiling *)
 writeln(logf,'PutPixel: 500000 calls');
 timerReset;
 for i:=1 to 500000 do
  PutPixel(random(640),random(480),random(256));
 writeln(logf,'BGI PutPixel: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 500000 do
  _PutPixel(random(640),random(480),random(256));
 writeln(logf,'_PutPixel: ',timerRead:8:3,' sec');
 ClearDevice;

(* Line Profiling *)
 writeln(logf,'Line: 30000 calls');
 timerReset;
 for i:=1 to 30000 do begin
  SetColor(random(256));
  Line(random(640),random(480),random(640),random(480));
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 30000 do
  _Line(random(640),random(480),random(640),random(480),random(256));
 writeln(logf,'_Line: ',timerRead:8:3,' sec');
 ClearDevice;

(* Masked line profiling *)
 writeln(logf,'Masked Line: 30000 calls');
 timerReset;
 for i:=1 to 30000 do begin
  SetColor(random(256));
  SetLineStyle(4,random(65535),1);
  Line(random(640),random(480),random(640),random(480));
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 30000 do
  _MaskedLine(random(640),random(480),random(640),random(480),random(256),random(65535));
 writeln(logf,'_MaskedLine: ',timerRead:8:3,' sec');
 ClearDevice;
 SetLineStyle(0,0,1);

(* Horizontal line profiling *)
 writeln(logf,'Horizontal Line: 50000 calls');
 timerReset;
 for i:=1 to 50000 do begin
  y1:=random(480);
  SetColor(random(256));
  Line(random(640),y1,random(640),y1);
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 50000 do begin
  x1:=random(640);
  x2:=random(640);
  if x1>x2 then xchange(x1,x2);
  _HLine(x1,x2,random(480),random(256));
 end;
 writeln(logf,'_HLine: ',timerRead:8:3,' sec');
 ClearDevice;

(* Vertical line profiling *)
 writeln(logf,'Vertical Line: 50000 calls');
 timerReset;
 for i:=1 to 50000 do begin
  x1:=random(640);
  SetColor(random(256));
  Line(x1,random(480),x1,random(480));
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 50000 do begin
  y1:=random(480);
  y2:=random(480);
  if y1>y2 then xchange(y1,y2);
  _VLine(random(640),y1,y2,random(256));
 end;
 writeln(logf,'_VLine: ',timerRead:8:3,' sec');
 ClearDevice;

(* PutImage profiling *)
GetMem(p1,65000);
GetMem(p2,65000);
for i:=0 to 199 do _HLine(0,319,i,i);
GetImage(0,0,319,199,p1^);
_GetImage(0,0,319,199,p2);
ClearDevice;
 writeln(logf,'PutImage: 5000 calls');
 timerReset;
 for i:=1 to 5000 do
  PutImage(random(320),random(280),p1^,NormalPut);
 writeln(logf,'BGI PutImage: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 5000 do
  _PutImage(random(320),random(280),320,200,p2);
 writeln(logf,'_PutImage: ',timerRead:8:3,' sec');
 ClearDevice;
 FreeMem(p1,65000);
 FreeMem(p2,65000);

(* PutText profiling *)
 _SetFont(@FONT);
 writeln(logf,'Put Text: 15000 calls');
 timerReset;
 for i:=1 to 15000 do begin
  SetColor(random(256));
  OutTextXY(random(560),random(456),'1234567890');
 end;
 writeln(logf,'BGI OutTextXY: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 15000 do
  _WriteStr(random(560),random(456),'1234567890',random(256));
 writeln(logf,'_WriteStr: ',timerRead:8:3,' sec');
 ClearDevice;

CloseGraph;
close(logf);
end.
