(********************************************)
(* VGA256.TPU vs. GRAPH.TPU: The Profiling  *)
(********************************************)
(* Example program for VRg SuperVGA library *)
(* Copyleft (C) 1996-1999 Virtual research  *)
(********************************************)

program vgaprof;
{$S-,Q-,R-,V-,B-,X+,G+,N+}
uses Graph,VGA256,CRT,DOS;
var driver:integer;
    ticks:longint;
    logf:text;
    x1,x2,y1,y2:word;
    i:longint;
    c,m:byte;
    mode:integer;
    p1,p2:pointer;
    f:file;
    Year,Month,Day,DayOfWeek:Word;
    Machine,VCard:string;

const days:array[0..6] of string[3]=('Sun','Mon','Tue','Wed','Thr','Fri','Sat');
      mont:array[1..12] of string[3]=
   ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');

procedure xchange(var a,b:word);
var c:word;
begin
 c:=a;
 a:=b;
 b:=c;
end;

procedure TimerReset;
begin
 ticks:=MemL[$0040:$006C];
end;

function TimerRead:real;
begin
 TimerRead:=(MemL[$0040:$006C]-ticks)/18.2;
end;

procedure FONT;external;
{$L FONT.OBJ}

procedure BGIVGA256;external;
{$L VGA256.OBJ}

begin
 clrscr;
 writeln (#10,#13,'VGA 320x200x256 graphics routines performance measurment tool');
 writeln('(C)oded by Asp / VR group, 1998-1999');
 writeln('This program will compare performance of VR group SuperVGA library');
 writeln('and Borland BGI graphics (VGA256.BGI).',#13,#10);
 writeln('Enter your machine description, please:');
 readln(Machine);
 writeln('Enter your videocard description, please:');
 readln(VCard);
 writeln(#10,#13,'Tests will take about 5..20 minutes depending on your hardware.');
 writeln('Tests results will be written to VGAPROF.LOG in current directory.');
 writeln('If there will be no screen activity for long time, please reboot your system.');
 writeln('Press a key to start tests now.',#13,#10);
 ReadKey;

 assign(logf,'vgaprof.log');
 rewrite(logf);
 GetDate(Year,Month,Day,DayOfWeek);
 writeln(logf,'VGAPROF log file');
 writeln(logf,'================');
 writeln(logf,Copyleft_Id);
 write(logf,'VGAPROF log file generated on ',days[DayOfWeek],' ',Day,'.',mont[Month],'.',Year,', ');
 GetTime(Year,Month,Day,DayOfWeek);
 writeln(logf,Year,':',Month,':',Day);
 writeln(logf,'Machine: '+Machine,#13,#10,'Video: '+VCard,#13,#10);
 driver:=InstallUserDriver('VGA256',nil);
 if RegisterBGIdriver(@BGIVGA256)<0 then begin
  writeln(#10,#13,'Cannot init graphics driver (EXE corrupted?)');
  halt(1);
 end;
 mode:=0;
 InitGraph(driver,mode,'');
 vgaRestoreDisplayBounds;
 randomize;

 (* Ellipse profiling *)
 writeln(logf,'Ellipse: 30000 calls');
 timerReset;
 for i:=1 to 30000 do begin
  SetColor(random(256));
  Ellipse(random(120)+100,100,0,360,random(100),random(100));
 end;
 writeln(logf,'BGI Ellipse: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 30000 do
  vgaEllipse(random(120)+100,100,random(100),random(100),random(256));
 writeln(logf,'vgaEllipse: ',timerRead:8:3,' sec');
 ClearDevice;

(* Rectangle profiling *)
 writeln(logf,'Rectangle: 60000 calls');
 timerReset;
 for i:=1 to 60000 do begin
  x1:=random(320);
  y1:=random(200);
  x2:=random(320);
  y2:=random(200);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  SetColor(random(256));
  Rectangle(x1,y1,x2,y2);
 end;
 writeln(logf,'BGI Rectangle: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 60000 do begin
  x1:=random(320);
  y1:=random(200);
  x2:=random(320);
  y2:=random(200);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  vgaBox(x1,y1,x2,y2,random(256),0);
 end;
 writeln(logf,'vgaBox: ',timerRead:8:3,' sec');
 ClearDevice;

(* Bar profiling *)
 writeln(logf,'Bar: 30000 calls');
 timerReset;
 for i:=1 to 30000 do begin
  x1:=random(320);
  y1:=random(200);
  x2:=random(320);
  y2:=random(200);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  SetFillStyle(1,random(256));
  Bar(x1,y1,x2,y2);
 end;
 writeln(logf,'BGI Bar: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 30000 do begin
  x1:=random(320);
  y1:=random(200);
  x2:=random(320);
  y2:=random(200);
  if x1>x2 then xchange(x1,x2);
  if y1>y2 then xchange(y1,y2);
  c:=random(255)+1;
  vgaBox(x1,y1,x2,y2,c,c);
 end;
 writeln(logf,'vgaBox: ',timerRead:8:3,' sec');
 ClearDevice;

(* PutPixel profiling *)
 writeln(logf,'PutPixel: 1000000 calls');
 timerReset;
 for i:=1 to 1000000 do 
  PutPixel(random(320),random(200),random(256));
 writeln(logf,'BGI PutPixel: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 1000000 do 
  vgaPutPixel(random(320),random(200),random(256));
 writeln(logf,'vgaPutPixel: ',timerRead:8:3,' sec');
 ClearDevice;

(* Line Profiling *)
 writeln(logf,'Line: 60000 calls');
 timerReset;
 for i:=1 to 60000 do begin
  SetColor(random(256));
  Line(random(320),random(200),random(320),random(200));
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 60000 do
  vgaLine(random(320),random(200),random(320),random(200),random(256));
 writeln(logf,'vgaLine: ',timerRead:8:3,' sec');
 ClearDevice;

(* Masked line profiling *)
 writeln(logf,'Masked Line: 60000 calls');
 timerReset;
 for i:=1 to 60000 do begin
  SetColor(random(256));
  SetLineStyle(4,random(65535),1);
  Line(random(320),random(200),random(320),random(200));
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 60000 do
  vgaMaskedLine(random(320),random(200),random(320),random(200),random(256),random(65535));
 writeln(logf,'vgaMaskedLine: ',timerRead:8:3,' sec');
 ClearDevice;
 SetLineStyle(0,0,1);

(* Horizontal line profiling *)
 writeln(logf,'Horizontal Line: 200000 calls');
 timerReset;
 for i:=1 to 200000 do begin
  y1:=random(200);
  c:=random(256);
  SetColor(random(256));
  Line(random(320),y1,random(320),y1);
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 200000 do begin
  x1:=random(320);
  x2:=random(320);
  if x1>x2 then xchange(x1,x2);
  c:=random(256);
  vgaHLine(x1,x2,random(200),random(256));
 end;
 writeln(logf,'vgaHLine: ',timerRead:8:3,' sec');
 ClearDevice;

(* Vertical line profiling *)
 writeln(logf,'Vertical Line: 200000 calls');
 timerReset;
 for i:=1 to 200000 do begin
  x1:=random(320);
  SetColor(random(256));
  Line(x1,random(200),x1,random(200));
 end;
 writeln(logf,'BGI Line: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 200000 do begin
  y1:=random(200);
  y2:=random(200);
  if y1>y2 then xchange(y1,y2);
  vgaVLine(random(320),y1,y2,random(256));
 end;
 writeln(logf,'vgaVLine: ',timerRead:8:3,' sec');
 ClearDevice;

(* PutImage profiling *)
 GetMem(p1,65000);
 GetMem(p2,65000);
 for i:=0 to 99 do vgaHLine(0,159,i,i);
 GetImage(0,0,159,99,p1^);
 vgaGetImage(0,0,159,99,p2);
 ClearDevice;
 writeln(logf,'PutImage: 15000 calls');
 timerReset;
 for i:=1 to 15000 do
  PutImage(random(160),random(100),p1^,NormalPut);
 writeln(logf,'BGI PutImage: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 15000 do
  vgaPutImage(random(160),random(100),160,100,p2);
 writeln(logf,'vgaPutImage: ',timerRead:8:3,' sec');
 ClearDevice;
 FreeMem(p1,65000);
 FreeMem(p2,65000);

(* PutText profiling *)
 writeln(logf,'Put Text: 30000 calls');
 vgaSetFont(@FONT);
 timerReset;
 for i:=1 to 30000 do begin
  SetColor(random(256));
  OutTextXY(random(240),random(184),'1234567890');
 end;
 writeln(logf,'BGI OutTextXY: ',timerRead:8:3,' sec');
 ClearDevice;
 timerReset;
 for i:=1 to 30000 do begin
  vgaWriteStr(random(240),random(184),'1234567890',random(256));
 end;
 writeln(logf,'vgaWriteStr: ',timerRead:8:3,' sec');
 ClearDevice;

CloseGraph;
close(logf);
end.

