(********************************************************)
(* Autodesk FLI animation player for Borland Pascal 7.0 *)
(* Copyleft (C) 1998-1999 Virtual Research i. g.        *)
(* All Lefts Reserved                                   *)
(*                                                      *)
(* Unit version : 1.21                                  *)
(* Started      : 30.May.1998                           *)
(* Finished     : 01.Jun.1998                           *)
(* Last updated : 16.Jul.1999                           *)
(* System       : DOS/DPMI16                            *)
(********************************************************)

unit FLI;
{$S-,R-,Q-,V-,B-,X+,G+}

interface

       (* Flags *)
 const Version='1.21';
       fliInfiniteLoop:boolean=false;   (* Loop flick for ever *)
       fliCheckKeyPress:boolean=true;   (* Exit on <Esc> *)
       fliLoopTimes:word=1;             (* Loop some times *)

       fliUserPreFrame:procedure=nil;	(* Will be called before each frame *)
       fliUserPostFrame:procedure=nil;	(* Will be called at the end of
					   each frame *)

 function  fliAllocateVFB:pointer;
 procedure fliFreeVFB;
 procedure fliPlayFromFile(FileName:string;FileOffset:longint);



implementation
uses VGA256,Mem32;

var Ticks:longint;

procedure TimerReset;
begin
 Ticks:=MemL[seg0040:$6C];
end;

function TimerRead:longint;
begin
 TimerRead:=MemL[seg0040:$6C]-Ticks;
end;

{$I flick.inc}

type T_VFB=array[0..63999] of byte;
     P_VFB=^T_VFB;

var   FLI_Header:TFlickHeader;
      VFB:P_VFB;                        (* Virtual framebuffer *)
      VFBPos:word;                      (* Relative position in VFB *)
      Palette:TPalette;
      Position:word;
      PFB:pointer;                 	(* Pointer to packed frame *)

const _FLI_COLOR=11;
      _FLI_LC=12;
      _FLI_BLACK=13;
      _FLI_BRUN=15;
      _FLI_COPY=16;

function  fliAllocateVFB:pointer;
begin
 GetMem(VFB,64000);
 fliAllocateVFB:=VFB;
end;

procedure fliFreeVFB;
begin
 FreeMem(VFB,64000);
 VFB:=nil;
end;

procedure GetData(var Destination;Count:word);
begin
 Move(ptr(seg(PFB^),ofs(PFB^)+Position)^,Destination,Count);
 Inc(Position,Count);
end;

procedure FLI_COLOR;
var Packets,i:word; Skip:byte;
const PalettePosition:byte=0; Count:word=0; Dummy:word=0;
begin
 GetData(Packets,2);
 for i:=1 to Packets do begin
  GetData(Skip,1);
  GetData(Count,1);
  Inc(Dummy,Count);
  if Count=0 then Count:=256;
  Inc(PalettePosition,Skip);
  GetData(Palette[PalettePosition],Count*3);
  Inc(PalettePosition,Count);
 end;
 if Odd(Dummy*3) then GetData(Skip,1);
 vgaSetPalette(Palette);
end;

procedure FLI_LC;
var SkipLines,ChangeLines,i:word; Packets,j:byte;
    SkipPix,PixData:byte; ChangePix:shortint;
begin
 VFBPos:=0;
 GetData(SkipLines,2);
 GetData(ChangeLines,2);
 for i:=1 to ChangeLines do begin
  VFBPos:=(SkipLines+i-1)*320;
  GetData(Packets,1);
  for j:=1 to Packets do begin
   GetData(SkipPix,1);
   Inc(VFBPos,SkipPix);
   GetData(ChangePix,1);
   if ChangePix>=0 then begin
    GetData(VFB^[VFBPos],ChangePix);
    Inc(VFBPos,ChangePix);
   end
   else begin
    GetData(PixData,1);
    FillChar(VFB^[VFBPos],-ChangePix,PixData);
    Inc(VFBPos,-ChangePix);
   end;
  end;
 end;
end;

procedure FLI_BLACK;
begin
 FillChar(VFB^,64000,0);
end;

procedure FLI_BRUN;
var i:word; Packets,j:byte;
    PixData:byte; ChangePix:shortint;
begin
 VFBPos:=0;
 for i:=1 to 200 do begin
  GetData(Packets,1);
  for j:=1 to Packets do begin
   GetData(ChangePix,1);
   if ChangePix<0 then begin
    GetData(VFB^[VFBPos],-ChangePix);
    Inc(VFBPos,-ChangePix);
   end
   else begin
    GetData(PixData,1);
    FillChar(VFB^[VFBPos],ChangePix,PixData);
    Inc(VFBPos,ChangePix);
   end;
  end;
 end;
end;

procedure FLI_COPY;
begin
 GetData(VFB^,64000);
end;


procedure fliPlayFromFile(FileName:string;FileOffset:longint);
var Loop,FrameCount,ChunkCount:word;
    FrameHeader:TFrameHeader;
    ChunkSize:longint; ChunkType:word;
    FlickFile:file; Delay:longint;

begin
 if VFB=nil then Exit;
 GetMem(PFB,65000);
 Assign(FlickFile,FileName);
 Reset(FlickFile,1); Seek(FlickFile,FileOffset);
 BlockRead(FlickFile,FLI_Header,128);
 if FLI_Header.Signature<>$AF11 then Exit; (* Not a FLI *)
 if (FLI_Header.Width<>320) or (FLI_Header.Height<>200) then Exit;
                                               (* Wrong size *)
 Delay:=FLI_Header.Speed shr 2;
 repeat
  for Loop:=1 to fliLoopTimes do begin
   Seek(FlickFile,FileOffset+128);
   for FrameCount:=1 to FLI_Header.Frames do begin
    TimerReset;
    BlockRead(FlickFile,FrameHeader,16);
    BlockRead(FlickFile,PFB^,FrameHeader.Size-16);
    Position:=0;
    if FrameHeader.ID=$F1FA then begin
     for ChunkCount:=1 to FrameHeader.Chunks do begin
      GetData(ChunkSize,4);
      GetData(ChunkType,2);
      Case ChunkType of
       _FLI_COLOR:FLI_COLOR;
       _FLI_LC:FLI_LC;
       _FLI_BLACK:FLI_BLACK;
       _FLI_BRUN:FLI_BRUN;
       _FLI_COPY:FLI_COPY;
      end;
      if @fliUserPreFrame<>nil then fliUserPreframe;
      if ChunkType<>_FLI_COLOR then Move(VFB^,ptr(segA000,0)^,64000);
     end;
    end;
    if fliCheckKeyPress then if Port[$60]=1 then begin
     MemW[Seg0040:$001C-2]:=MemW[Seg0040:$001C]; (* Flush keyboard *)
     Exit;
    end;
    if @fliUserPostFrame<>nil then fliUserPostFrame;
    while TimerRead<Delay do;
   end;
  end;
 until (not fliInfiniteLoop);
 FreeMem(PFB,65000);
 Close(FlickFile);
end;

begin
 VFB:=nil;
end.
