(************************************************************************)
(* Super VGA library for Borland/Turbo Pascal 7.0                       *)
(* (C)reated by Asp / VR group, 1996-99                                 *)
(* Virtual Research independent group production, 1999                  *)
(*                                                                      *)
(* Unit version: 4.04                                                   *)
(* Last updated: 01.Jul.1999                                            *)
(* System      : DOS                                                    *)
(*                                                                      *)
(* portionally based on sources by Guy Desrosiers, Yaron Keren,         *)
(* Jos Walbeehm, Peter van Sebille, Christopher L. Tumber,              *)
(* Jan van Loenen, Kosmas Einbrodt, Oliver Fromme, and David Hackamack  *)
(* Special thanx to: E. Shikin and A. Boreskov;                         *)
(*                   SWAG support team and SWAG contributors;           *)
(*                   all the Pascal programmers allover the world :)    *)
(*                                                                      *)
(* This unlicensed and copylefted material is protected by              *)
(* the Second Law of Thermodynamics. All lefts reserved.                *)
(************************************************************************)

unit SVGA;
{$X+,N+,E+,G+,F+,S-,Q-,R-,V-,B-,O-}

interface

 const Version='4.04';

 type T_VESAInfo=record                 (* VESA VBE information *)
       Signature:array[0..3] of char;   (* 'VESA' if VBE detected *)
       Version:word;                    (* BCD version *)
       OEMId:pointer;       (* Pointer to OEM string *)
       Flags:longint;       (* Hardware capabilities *)
       ModeList:pointer;    (* Pointer to modes list. You should not use it *)
       VideoMemory:word;    (* Videomemory amount in 64 Kb blocks *)
       Reserved:array[1..236] of byte;
      end;

      TRGB=record
       R,G,B:byte
      end;

      TPalette=array[0..255] of TRGB;

 var _MaxX,_MaxY:word;
     _MinClipX,_MinClipY,_MaxClipX,_MaxClipY:word;  (* Clipwindow *)
     _Mode:word;                     (* Current mode *)
     _VESAPresent:boolean;           (* True if VESA detected *)
     _VESAInfo:T_VESAInfo;           (* Information returned by VBE *)
     _Chipset:word;                  (* Selected chipset *)
     _TransparentColor:byte;         (* "Transparent" color *)

 const _640x350:word=0;              (* Available modes *)
       _640x400:word=0;
       _640x480:word=0;
       _800x600:word=0;
       _1024x768:word=0;

                        (* Chipsets *)

 const _VESA=0;              (* VESA compatible, default *)
       _AHEADA=1;            (* Ahead A *)
       _AHEADB=2;            (* Ahead B *)
       _ATI=3;               (* ATI (not tested and may not work) *)
       _CHIPSNTECS=4;        (* Chips & Technologies  82C45x *)
       _CIRRUS1M=5;          (* Cirrus Logic CL-GD 542x with 1 M *)
       _CIRRUS2M=6;          (* Cirrus Logic CL-GD 542x with 2 M *)
       _COMPAQ=7;            (* Compaq QVision/AVGA *)
       _ET3000=8;            (* Tseng Labs ET3000 *)
       _ET4000=9;            (* Tseng Labs ET4000 *)
       _ET4000W32=10;        (* Tseng ET4000-W32 *)
       _GENOA=11;            (* Genoa 6x00 chipset *)
       _NCR=12;              (* NCR 77C22 *)
       _OAK=13;              (* OAK *)
       _PARADISE=14;         (* Paradise / Western Digital WD90Cxx *)
       _REALTEK=15;          (* Realtek *)
       _TRIDENT=16;          (* Trident 8900 and 9000 *)
       _VIDEO7=17;           (* Video 7  VRAM II *)


 procedure _Init(chipset:byte);
 function  _SetMode(mode:word):boolean;
 procedure _InitMode(mode:word);

 procedure _PutPixel(x,y:word;color:byte);
 procedure _PutTransparentPixel(x,y:word;color:byte);
 procedure _GetPixel(x,y:word;var color:byte);
 procedure _Line(x1,y1,x2,y2:word;color:byte);
 procedure _MaskedLine(x1,y1,x2,y2:word;color:byte;mask:word);
 procedure _Ellipse(xc,yc,rx,ry:word;color:byte);
 procedure _FilledEllipse(xc,yc,rx,ry:word;color:byte);
 procedure _Box(x1,y1,x2,y2:word;color,fill:byte);
 procedure _HLine(x1,x2,y:word;color:byte);
 procedure _VLine(x,y1,y2:word;color:byte);
 procedure _Triangle(x1,y1,x2,y2,x3,y3:word;color:byte);
 procedure _FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
 procedure _FilledPolygon(var VertexArray;NumOfVertexes,color:byte);

 procedure _SetPalette(var pal);
 procedure _FixedPalette(var pal);
 procedure _GetPalette(var pal);
 procedure _SetColor(color,r,g,b:byte);
 procedure _GetColor(color:byte;var r,g,b:byte);
 procedure _FadeIn(pal:array of byte;speed:single);
 procedure _FadeOut(pal:array of byte;speed:single);

 procedure _Fill(color:byte);
 procedure _SetTextMode;
 procedure _WaitVSync;
 procedure _WaitVSyncStart;
 procedure _RestoreDisplayBounds;

 procedure _LoadPCX(filename:string;fileoffset:longint;x,y:word);
 procedure _GetPCXPalette(filename:string;var palette);
 procedure _SavePCX(filename:string;x1,y1,x2,y2:word);
 procedure _GetPCXSize(filename:string;fileoffset:longint;var x,y:word);

 procedure _GetImage(x1,y1,x2,y2:word;clip:pointer);
 procedure _PutImage(x1,y1:integer;width,height:word;clip:pointer);
 procedure _XORImage(x1,y1:integer;width,height:word;clip:pointer);
 procedure _PutTransparentImage(x1,y1:integer;width,height:word;clip:pointer);
 procedure _GetLargeImage(x1,y1,x2,y2:word;handle:word);
 procedure _PutLargeImage(x1,y1:integer;width,height:word;handle:word);

 procedure _WriteStr(x,y:word;S:string;color:byte);
 procedure _SetFont(p:pointer);

 procedure __PutPixel(x,y:integer;color:byte);
 procedure __PutTransparentPixel(x,y:integer;color:byte);
 procedure __GetPixel(x,y:integer;var color:byte);
 procedure __Line(x1,y1,x2,y2:integer;color:byte);
 procedure __MaskedLine(x1,y1,x2,y2:integer;color:byte;mask:word);
 procedure __Ellipse(xc,yc:integer;rx,ry:word;color:byte);
 procedure __FilledEllipse(xc,yc:integer;rx,ry:word;color:byte);
 procedure __Box(x1,y1,x2,y2:integer;color,fill:byte);
 procedure __HLine(x1,x2,y:integer;color:byte);
 procedure __VLine(x,y1,y2:integer;color:byte);
 procedure __Triangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
 procedure __FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
 procedure __FilledPolygon(var VertexArray;NumOfVertexes,color:byte);

 procedure _SetLogicalWidth(newwidth:word);
 procedure _SetDisplayOrigin(x,y:word);

 function  Copyleft_Id:string;


implementation
uses XManager,Mem32;

var Granularity:byte;                   (* Window granularity *)
    CurBank:word;                       (* Current bank *)
    CnTchipset:byte;                    (* Chips'n'Techs id *)
    XMDescriptor:TxmmMoveStruc;         (* Structure to swap data <-> XMS *)
    FontPtr:pointer;                    (* Current font *)

    SwitchBank:procedure(bank:word);    (* Switch bank routine *)
    VESA_WinCtl:longint;          (* VESA window control procedure address *)
    _InitGraph:procedure;         (* Set up svga card routine *)
    _ShutDown:procedure;          (* Set down svga card routine *)

    Copyleft:string;                    (* Product information *)


procedure FindVESA;
        (* Finds if VESA VBE is present *)
label Exit_;
begin
 asm
  push ds
  pop es
  mov di,offset _VESAInfo
  mov ax,$4F00
  int $10
  cmp ax,$004F
  je @@1
  mov _VESAPresent,false
  jmp Exit_
@@1:
 end;
 if _VESAInfo.Signature<>'VESA' then _VESAPresent:=false
  else _VESAPresent:=true;
Exit_:
end;

(***************************************************************************

                         Switch bank procedures

****************************************************************************)

(* Each one of these must preserve cx, as it is assumed by other routines *)

procedure VESASwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 cmp Granularity,1
 je @@1
 mul Granularity
@@1:
 mov dx,ax
 mov ax,$4F05           (* We do not use double-windowing here *)
 xor bx,bx
 call [VESA_WinCtl]
end;

procedure AHEADASwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 push ax
 mov dx,$3CE
 mov ax,$200F
 out dx,ax
 pop ax
 push ax
 push bx
 mov bl,al
 shl ax,7
 mov al,0dh
 out dx,ax
 mov dl,$CC
 in al,dx
 mov dl,0c2h
 and al,11011111b
 and bl,01h
 shl bl,5
 or al,bl
 out dx,al
 pop bx
 pop ax
end;

procedure AHEADBSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 push ax
 mov dx,$03CE
 mov ax,$200F
 out dx,ax
 pop ax
 push ax
 mov ah,al
 shl ah,4
 or ah,al
 mov al,$0D
 out dx,ax
 pop ax
end;

procedure ATISwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 mov dl,al
 shl al,4
 or al,dl
 rol al,1
 push ax
 mov dx,$01CE
 mov al,$B2
 out dx,al
 inc dx
 pop ax
 out dx,al
end;

procedure CHIPSNTECSSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 push ax
 mov dl,CnTchipset
 test dl,dl
 jz @@Bank0
 dec dl
 jz @@Bank1
@@Bank2:
 mov dx,$03D6
 shl ax,12
 mov al,$10
 out dx,ax
 pop ax
 jmp @@Exit
@@Bank1:
 mov dx,$03D6
 shl ax,10
 mov al,$10
 out dx,ax
 pop ax
 jmp @@Exit
@@Bank0:
 mov dx,$03D6
 mov ah,al
 mov al,$0B
 out dx,ax
 pop ax
@@Exit:
end;

procedure CIRRUS1MSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 shl al,4
 push ax
 mov dx,$03CE
 mov al,9
 out dx,al
 inc dx
 pop ax
 out dx,al
 dec dx
 mov al,8
 out dx,al
end;

procedure CIRRUS2MSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 shl al,2
 push ax
 mov dx,$03CE
 mov al,9
 out dx,al
 inc dx
 pop ax
 out dx,al
 dec dx
 mov al,8
 out dx,al
end;

procedure COMPAQSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 shl al,4
 push ax
 mov dx,$03CE
 mov al,$45
 out dx,al
 inc dx
 pop ax
 push ax
 out dx,al
 dec dx
 mov al,46h
 out dx,al
 inc dx
 pop ax
 out dx,al
end;

procedure ET3000SwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 mov dl,al
 shl al,3
 or al,dl
 or al,$40
 mov dx,$03CD
 out dx,al
end;

procedure ET4000SwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 mov dl,al
 shl al,4
 or  al,dl
 mov dx,$03CD
 out dx,al
end;

procedure ET4000W32SwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 push ax
 and al,$0F
 mov dl,al
 shl al,4
 or  al,dl
 mov dx,$03CD
 out dx,al
 pop ax
 and al,$F0
 mov dl,al
 shr al,4
 or al,dl
 mov dx,$03CB
 out dx,al
end;

procedure GENOASwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 mov dl,al
 shl al,3
 or al,dl
 or al,$40
 push ax
 mov al,$06
 mov dx,$03C4
 out dx,al
 pop ax
 inc dx
 out dx,al
end;

procedure NCRSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 shl al,2
 push ax
 mov dx,$03C4
 mov al,$18
 out dx,al
 inc dx
 pop ax
 out dx,al
end;

procedure OAKSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 mov dl,al
 shl al,4
 or  al,dl
 push ax
 mov dx,$03DE
 mov al,$11
 out dx,al
 inc dx
 pop ax
 out dx,al
end;

procedure PARADISESwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 shl al,4
 push ax
 mov dx,$03CE
 mov al,$09
 out dx,al
 inc dx
 pop ax
 push ax
 out dx,al
 dec dx
 mov al,$0A
 out dx,al
 inc dx
 pop ax
 out dx,al
end;

procedure REALTEKSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 mov dx,$3D7
 out dx,al
end;

procedure TRIDENTSwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 push ax
 mov al,$0E
 mov dx,$03C4
 out dx,al
 inc dx
 pop ax
 xor al,$02
 out dx,al
end;

procedure VIDEO7SwitchBank(bank:word);assembler;
asm
 mov ax,bank
 mov CurBank,ax
 push ax
 shl al,4
 mov ah,al
 mov al,$E8
 mov dx,$03C4
 out dx,ax
 pop ax
end;


(***************************************************************************

                         InitGraph procedures

****************************************************************************)

procedure ATIInitGraph;assembler;
asm
 mov dx,$01CE
 mov al,$BE
 out dx,al
 inc dx
 in al,dx
 or al,8
 out dx,al
end;

procedure CHIPSNTECSInitGraph;assembler;
asm
 mov dx,$46E8
 mov ax,$001E
 out dx,ax           (* enter setup mode *)
 push dx
 mov dx,$0103
 mov al,$80
 out dx,ax           (* enable extended registers *)
 pop dx
 mov al,$0E
 out dx,ax           (* leave setup mode *)
 mov dx,$03D6
 mov al,4
 out dx,al
 inc dx
 in al,dx
 or al,4
 out dx,al           (* enable bank access & extended CRTC *)
 dec dx
 mov al,$0B
 out dx,al
 inc dx
 in al,dx
 and al,$FC
 or al,1
 out dx,al           (* enable extended paging & single page mode *)
 dec dx
 xor ax,ax
 out dx,al
 inc dx
 in al,dx
 shr al,4
 jz @@ready   (* 82c451 -> 0 *)
 dec al
 jz @@ready1  (* 82c452 -> 1 *)
 dec al
 jz @@ready   (* 82c455 -> 0 *)
 cmp al,3
 jb @@ready2  (* 82c453/0 -> 2 *)
 cmp al,5
 jb @@ready   (* 82c456/7 -> 0 *)
              (* F655x0 -> 2 *)
@@ready2:
 inc ah
@@ready1:
 inc ah
@@ready:
 mov CnTchipset,ah
end;

procedure CIRRUS1MInitGraph;assembler;
asm
 mov dx,$03C4
 mov al,6
 out dx,al
 inc dx
 mov al,$12
 out dx,al
end;

procedure CIRRUS2MInitGraph;assembler;
asm
 mov dx,$03C4
 mov al,6
 out dx,al
 inc dx
 mov al,$12
 out dx,al
 mov dl,$CE
 mov al,$0B
 out dx,al
 inc dx
 in al,dx
 or al,$20
 out dx,al
end;

procedure COMPAQInitGraph;assembler;
asm
 mov dx,$03CE
 mov al,$0F
 out dx,al
 inc dx
 mov al,5
 out dx,al   (* unlock extended registers *)
 dec dx
 mov al,$40
 out dx,al
 inc dx
 in al,dx
 push ax
 dec dx
 mov al,$40
 out dx,al
 inc dx
 pop ax
 and al,$FE
 out dx,al   (* page 1 relative to a000:0 *)
end;

procedure ET3000InitGraph;assembler;
asm
 mov dx,$03BF
 mov al,3
 out dx,al
 mov dl,$D8
 mov al,$A0
 out dx,al
end;

procedure PARADISEInitGraph;assembler;
asm
 mov dx,$03CE
 mov al,$0F
 out dx,al
 inc dx
 mov al,5
 out dx,al
 mov dx,$03C4
 mov al,6
 out dx,al
 inc dx
 mov al,$48
 out dx,al
 mov dx,$03D4
 mov al,$29
 out dx,al
 inc dx
 mov al,$85
 out dx,al
 dec dx
 mov al,$2F
 out dx,al
 inc dx
 in al,dx
 and al,11111101b
 out dx,al
 mov dx,$03CE
 mov al,$0B
 out dx,al
 inc dx
 in al,dx
 or al,00001000b
 out dx,al
 mov dx,$03C4
 mov al,$11
 out dx,al
 inc dx
 in al,dx
 or al,10000000b
 out dx,al
end;

procedure TRIDENTInitGraph;assembler;
asm
 mov dx,$3CE         (* set pagesize to 64k *)
 mov al,6
 out dx,al
 inc dl
 in al,dx
 dec dl
 or al,4
 mov ah,al
 mov al,6
 out dx,ax
 mov dl,$C4         (* switch to BPS mode *)
 mov al,$0B
 out dx,al
 inc dl
 in al,dx
end;

procedure VIDEO7InitGraph;assembler;
asm
 mov dx,$03C4
 mov al,6
 out dx,al
 inc dx
 mov al,$EA
 out dx,al
end;

(***************************************************************************

                         ShutDown procedures

****************************************************************************)

procedure PARADISEShutDown;assembler;
asm
 mov dx,$03CE
 mov al,$0F
 out dx,al
 inc dx
 xor al,al
 out dx,al
 dec dx
 mov al,6
 out dx,al
 inc dx
 xor al,al
 out dx,al
end;

(***************************************************************************

                         Init and Set mode procedures

****************************************************************************)

procedure _Init(chipset:byte);
        (* Initialize graphics system *)
begin
 case chipset of
  _VESA:begin
              _640x350:=$11C;
              _640x400:=$100;
              _640x480:=$101;
              _800x600:=$103;
              _1024x768:=$105;
              @SwitchBank:=@VESASwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _AHEADA:begin
              _640x350:=0;
              _640x400:=$60;
              _640x480:=$61;
              _800x600:=$62;
              _1024x768:=0;
              @SwitchBank:=@AHEADASwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _AHEADB:begin
              _640x350:=0;
              _640x400:=$60;
              _640x480:=$61;
              _800x600:=$62;
              _1024x768:=$63;
              @SwitchBank:=@AHEADBSwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _ATI:begin
              _640x350:=0;
              _640x400:=$61;
              _640x480:=$62;
              _800x600:=$63;
              _1024x768:=$64;
              @SwitchBank:=@ATISwitchBank;
              @_InitGraph:=@ATIInitGraph;
              @_ShutDown:=nil;
             end;
  _CHIPSNTECS:begin
              _640x350:=0;
              _640x400:=$78;
              _640x480:=$79;
              _800x600:=$7B;
              _1024x768:=$7E;
              @SwitchBank:=@CHIPSNTECSSwitchBank;
              @_InitGraph:=@CHIPSNTECSInitGraph;
              @_ShutDown:=nil;
             end;
  _CIRRUS1M:begin
              _640x350:=0;
              _640x400:=0;
              _640x480:=$5F;
              _800x600:=$5C;
              _1024x768:=$60;
              @SwitchBank:=@CIRRUS1MSwitchBank;
              @_InitGraph:=@CIRRUS1MInitGraph;
              @_ShutDown:=nil;
             end;
  _CIRRUS2M:begin
              _640x350:=0;
              _640x400:=0;
              _640x480:=$5F;
              _800x600:=$5C;
              _1024x768:=$60;
              @SwitchBank:=@CIRRUS2MSwitchBank;
              @_InitGraph:=@CIRRUS2MInitGraph;
              @_ShutDown:=nil;
             end;
  _COMPAQ:begin
              _640x350:=0;
              _640x400:=0;
              _640x480:=$32;
              _800x600:=$34;
              _1024x768:=$38;
              @SwitchBank:=@COMPAQSwitchBank;
              @_InitGraph:=@COMPAQInitGraph;
              @_ShutDown:=nil;
             end;
  _ET3000:begin
              _640x350:=$2D;
              _640x400:=0;
              _640x480:=$2E;
              _800x600:=$30;
              _1024x768:=0;
              @SwitchBank:=@ET3000SwitchBank;
              @_InitGraph:=@ET3000InitGraph;
              @_ShutDown:=nil;
             end;
  _ET4000:begin
              _640x350:=$2D;
              _640x400:=$2F;
              _640x480:=$2E;
              _800x600:=$30;
              _1024x768:=$38;
              @SwitchBank:=@ET4000SwitchBank;
              @_InitGraph:=@ET3000InitGraph;
              @_ShutDown:=nil;
             end;
  _ET4000W32:begin
              _640x350:=$2D;
              _640x400:=$2F;
              _640x480:=$2E;
              _800x600:=$30;
              _1024x768:=$38;
              @SwitchBank:=@ET4000W32SwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _GENOA:begin
              _640x350:=$5B;
              _640x400:=$7E;
              _640x480:=$2E;
              _800x600:=$30;
              _1024x768:=$38;
              @SwitchBank:=@GENOASwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _NCR:begin
              _640x350:=0;
              _640x400:=$5E;
              _640x480:=$5F;
              _800x600:=$5C;
              _1024x768:=$61;
              @SwitchBank:=@NCRSwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _OAK:begin
              _640x350:=0;
              _640x400:=0;
              _640x480:=$53;
              _800x600:=$54;
              _1024x768:=$59;
              @SwitchBank:=@OAKSwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _PARADISE:begin
              _640x350:=0;
              _640x400:=$5E;
              _640x480:=$5F;
              _800x600:=$5C;
              _1024x768:=$60;
              @SwitchBank:=@PARADISESwitchBank;
              @_InitGraph:=@PARADISEInitGraph;
              @_ShutDown:=@PARADISEShutDown;
             end;
  _REALTEK:begin
              _640x350:=0;
              _640x400:=$25;
              _640x480:=$26;
              _800x600:=$27;
              _1024x768:=$28;
              @SwitchBank:=@REALTEKSwitchBank;
              @_InitGraph:=nil;
              @_ShutDown:=nil;
             end;
  _TRIDENT:begin
              _640x350:=0;
              _640x400:=0;
              _640x480:=$5D;
              _800x600:=$5E;
              _1024x768:=$62;
              @SwitchBank:=@TRIDENTSwitchBank;
              @_InitGraph:=@TRIDENTInitGraph;
              @_ShutDown:=nil;
             end;
  _VIDEO7:begin
              _640x350:=0;
              _640x400:=0;
              _640x480:=$67;
              _800x600:=$69;
              _1024x768:=$6A;
              @SwitchBank:=@VIDEO7SwitchBank;
              @_InitGraph:=@VIDEO7InitGraph;
              @_ShutDown:=nil;
             end;
 end;
 _Chipset:=chipset;
 _Mode:=0;
 FontPtr:=nil;
 CurBank:=0;
end;

procedure GetModeInfo(mode:word);
        (* Get VESA mode granularity and window control function *)
var ModeInfo:array [1..128] of word;
    ModeInfoPtr:pointer;
begin
 ModeInfoPtr:=@Modeinfo;
 asm
  les di,ModeInfoPtr
  mov cx,mode
  mov ax,$4F01
  int $10
 end;
 Granularity:=64 div ModeInfo[3];
 VESA_WinCtl:=(longint(ModeInfo[8]) shl 16)+ModeInfo[7];
end;

function _SetMode(mode:word):boolean;
        (* Set svga videomode *)
var success:boolean;
begin
 if mode=0 then begin
  _SetMode:=false;
  Exit;
 end;
 if (_Mode<>0) and (@_ShutDown<>nil) then _ShutDown;
 _Mode:=mode;
 if _Chipset=_VESA then begin
  GetModeInfo(mode);
  asm
   mov ax,$4F02
   mov bx,mode
   int $10
   cmp ax,$004F
   je @@1
   mov success,false
   jmp @@2
@@1:
   mov success,true
@@2:
  end;
 end
 else begin
  asm
   mov ax,mode
   int $10
   mov ax,$0F00
   int $10
   and ax,$00FF
   cmp ax,mode
   je @@3
   mov success,false
   jmp @@4
@@3:
   mov success,true
@@4:
  end;
 end;
 _SetMode:=success;
 if not success then Exit;
 if @_InitGraph<>nil then _InitGraph;
 _TransparentColor:=0;
 _MinClipX:=0;_MinClipY:=0;
 if mode=_640x350 then begin
  _MaxX:=639;
  _MaxY:=349;
 end else
 if mode=_640x400 then begin
  _MaxX:=639;
  _MaxY:=399;
 end else
 if mode=_640x480 then begin
  _MaxX:=639;
  _MaxY:=479;
 end else
 if mode=_800x600 then begin
  _MaxX:=799;
  _MaxY:=599;
 end else
 if mode=_1024x768 then begin
  _MaxX:=1023;
  _MaxY:=767;
 end;
 _MaxClipX:=_MaxX;_MaxClipY:=_MaxY;
 CurBank:=0;
end;

procedure _InitMode(mode:word);
        (* Set up variables, but do not set videomode, *)
        (* for compatibility with BGI *)
begin
 if mode=0 then Exit;
 _Mode:=mode;
 if _Chipset=_VESA then GetModeInfo(mode);
 _MinClipX:=0; _MinClipY:=0;
 if mode=_640x350 then begin
  _MaxX:=639;
  _MaxY:=349;
 end else
 if mode=_640x400 then begin
  _MaxX:=639;
  _MaxY:=399;
 end else
 if mode=_640x480 then begin
  _MaxX:=639;
  _MaxY:=479;
 end else
 if mode=_800x600 then begin
  _MaxX:=799;
  _MaxY:=599;
 end else
 if mode=_1024x768 then begin
  _MaxX:=1023;
  _MaxY:=767;
 end;
 _MaxClipX:=_MaxX;_MaxClipY:=_MaxY;
 CurBank:=0;
end;

(***************************************************************************

                        PutPixel and GetPixel ones

****************************************************************************)

procedure _PutPixel(x,y:word;color:byte);assembler;
asm
 mov ax,$A000
 mov es,ax
 mov ax,_MaxX
 inc ax
 mul y
 add ax,x
 adc dx,0               (* Now bank is in DX, offset in AX *)
 mov di,ax
 cmp dx,CurBank
 je @@1
 push dx
 call SwitchBank
@@1:
 mov al,color
 mov es:[di],al
end;

procedure _PutTransparentPixel(x,y:word;color:byte);assembler;
 (* This version of PutPixel will not display
    color, if it's equal to _TransparentColor *)
asm
 mov al,_TransparentColor
 cmp al,color
 je @Exit
 mov ax,$A000
 mov es,ax
 mov ax,_MaxX
 inc ax
 mul y
 add ax,x
 adc dx,0               (* Now bank is in DX, offset in AX *)
 mov di,ax
 cmp dx,CurBank
 je @@1
 push dx
 call SwitchBank
@@1:
 mov al,color
 mov es:[di],al
 @Exit:
end;

procedure _GetPixel(x,y:word;var color:byte);assembler;
asm
 mov ax,$A000
 mov es,ax
 mov ax,_MaxX
 inc ax
 mul y
 add ax,x
 adc dx,0
 mov di,ax
 cmp dx,CurBank
 je @@1
 push dx
 call SwitchBank
@@1:
 mov al,es:[di]
 les di,color
 mov es:[di],al
end;

(***************************************************************************

                       PutPixel and GetPixel w/ clipping

****************************************************************************)

procedure __PutPixel(x,y:integer;color:byte);
begin
 if (x>=_MinClipX) and (x<=_MaxClipX) and
    (y>=_MinClipY) and (y<=_MaxClipY) then
  _PutPixel(x,y,color);
end;

procedure __PutTransparentPixel(x,y:integer;color:byte);
begin
 if (x>=_MinClipX) and (x<=_MaxClipX) and
    (y>=_MinClipY) and (y<=_MaxClipY) then
  _PutTransparentPixel(x,y,Color);
end;

procedure __GetPixel(x,y:integer;var color:byte);
begin
 if (x>=_MinClipX) and (x<=_MaxClipX) and
    (y>=_MinClipY) and (y<=_MaxClipY) then
  _GetPixel(x,y,color);
end;

(***************************************************************************

                               Palette routines

****************************************************************************)

procedure _SetPalette(var pal); assembler;
        (* Set palette from pal *)
asm
 mov dx,$3C8
 xor al,al
 out dx,al
 push ds
 lds si,pal
 cld
 inc dx
 mov cx,768
 rep outsb
 pop ds
end;

procedure _SetColor(color,r,g,b:byte);assembler;
asm
 mov dx,$3C8
 mov al,color
 out dx,al
 inc dx
 mov al,r
 out dx,al
 mov al,g
 out dx,al
 mov al,b
 out dx,al
end;

procedure _GetColor(color:byte;var r,g,b:byte);assembler;
asm
 mov dx,$3C7
 mov al,color
 out dx,al
 inc dx
 inc dx
 les di,r
 in al,dx
 mov es:[di],al
 les di,g
 in al,dx
 mov es:[di],al
 les di,b
 in al,dx
 mov es:[di],al
end;

procedure _FixedPalette(var pal);assembler;
        (* This one waits for vertical retrace before set palette *)
asm
 mov dx,$3DA
@@1:
 in al,dx
 test al,8
 jz @@1
 mov dx,$3C8
 xor al,al
 out dx,al
 push ds
 lds si,pal
 cld
 inc dx
 mov cx,768
 rep outsb
 pop ds
end;

procedure _GetPalette(var pal); assembler;
        (* Get current palette to pal *)
asm
 mov dx,$3C7
 xor al,al
 out dx,al
 les di,pal
 cld
 inc dx
 inc dx
 mov cx,768
 rep insb
end;

        (* The following two need big stack! *)

procedure _FadeIn(pal:array of byte;speed:single);
        (* Fade in to pal *)
var delta:array [0..767] of single;
    realpal:array [0..767] of single;
    i,j:word;
begin
 for i:=0 to 767 do begin
  realpal[i]:=pal[i];           (* make palette floating *)
  delta[i]:=realpal[i]/speed;
  realpal[i]:=0;
 end;
 for i:=1 to trunc(speed) do begin
  for j:=0 to 767 do begin
   realpal[j]:=realpal[j]+delta[j];
   pal[j]:=round(realpal[j]);
  end;
  _FixedPalette(pal[0]);
 end;
end;

procedure _FadeOut(pal:array of byte;speed:single);
        (* Fade out from pal *)
var delta:array [0..767] of single;
    realpal:array [0..767] of single;
    i,j:word;
begin
 for i:=0 to 767 do begin
  realpal[i]:=pal[i];           (* make palette floating *)
  delta[i]:=realpal[i]/speed;
 end;
 for i:=1 to trunc(speed) do begin
  for j:=0 to 767 do begin
   realpal[j]:=realpal[j]-delta[j];
   pal[j]:=trunc(realpal[j]);
  end;
  _FixedPalette(pal[0]);
 end;
end;

(***************************************************************************

                         Drawing routines

****************************************************************************)

procedure _HLine(x1,x2,y:word;color:byte);assembler;
        (* Fast horizontal line *)
var bank,ofset,count:word;
asm
 cld
 mov ax,$A000
 mov es,ax
 mov ax,x2
 sub ax,x1
 inc ax
 mov count,ax
 mov ax,_MaxX
 inc ax
 mul y
 add ax,x1
 adc dx,0
 mov bank,dx
 mov ofset,ax
 cmp dx,CurBank
 je @@1
 push bank
 mov cx,ax
 call SwitchBank
 mov ax,cx
@@1:
 add ax,count
 jc @@2                (* if next bank, goto @@2 *)
 mov di,ofset          (* if no, fill line *)
 mov cx,count
 mov al,color
 mov ah,al
 mov bx,ax
 db $66; shl ax,16     (* shl eax,16 *)
 mov ax,bx
 mov bx,cx
 shr cx,2
 db 66h; rep stosw     (* rep stosd *)
 and bx,3
 mov cx,bx
 rep stosb
 jmp @@Exit
@@2:
 mov di,ofset          (* fill part in current bank *)
 or cx,$FFFF           (* mov cx,$FFFF *)
 sub cx,di
 inc cx
 mov al,color
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 mov bx,cx
 shr cx,2
 db 66h; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
 inc bank
 push bank             (* switch to next bank *)
 call SwitchBank
 xor di,di             (* and fill the rest of horizontal line *)
 mov cx,ofset
 add cx,count
 mov al,color
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 mov bx,cx
 shr cx,2
 db 66h; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
@@Exit:
end;

procedure _VLine(x,y1,y2:word;color:byte);assembler;
        (* Fast vertical line *)
var bank,ofset:word;
asm
 mov ax,$A000           (* ES:0 points to videobuffer *)
 mov es,ax
 mov ax,_MaxX
 inc ax
 mul y1
 add ax,x
 adc dx,0
 mov bank,dx            (* bank and offset of start pixel *)
 mov ofset,ax
 cmp dx,CurBank         (* in current bank? *)
 je @@1
 push bank              (* if no, switch bank *)
 call SwitchBank
@@1:
 mov di,ofset
 mov al,color
 stosb                  (* put pixel *)

 mov cx,y2              (* define number of cycles *)
 sub cx,y1
 jcxz @@Exit
@@3:
 add di,_MaxX           (* offset of next pixel *)
 jnc @@2                (* if still in current bank, then put pixel *)
 inc bank
 push bank
 call SwitchBank
 mov al,color
@@2:
 stosb
 loop @@3
@@Exit:
end;

procedure _Line(x1,y1,x2,y2:word;color:byte);
        (* This is Bresenham's line.
           Still awaits to be rewritten in assembler :) *)
var d,dx,dy,ai,bi,xi,yi:integer;
begin
  if x1=x2 then begin
   if y1>y2 then begin d:=y2; y2:=y1; y1:=d; end;
   _VLine(x1,y1,y2,color);
   Exit;
  end;
  if y1=y2 then begin
   if x1>x2 then begin d:=x2; x2:=x1; x1:=d; end;
   _HLine(x1,x2,y1,color);
   Exit;
  end;
  if x1<x2 then begin xi:=1; dx:=x2-x1; end
  else begin xi:=-1; dx:=x1-x2; end;
  if y1<y2 then begin yi:=1; dy:=y2-y1; end
  else begin yi:=-1; dy:=y1-y2; end;
  asm
   mov ax,$A000
   mov es,ax
   mov ax,_MaxX
   inc ax
   mul y1
   add ax,x1
   adc dx,0
   mov di,ax
   cmp dx,CurBank
   je @@0
   push dx
   call SwitchBank
@@0:
   mov al,color
   mov es:[di],al
  end;
  if dx>dy then begin
   ai:=(dy-dx) shl 1;
   bi:=dy shl 1;
   d:=bi-dx;
   repeat
    if d>=0 then begin inc(y1,yi); inc(d,ai); end
    else inc(d,bi); inc(x1,xi);
    asm
     mov ax,$A000
     mov es,ax
     mov ax,_MaxX
     inc ax
     mul y1
     add ax,x1
     adc dx,0
     mov di,ax
     cmp dx,CurBank
     je @@1
     push dx
     call SwitchBank
@@1:
     mov al,color
     mov es:[di],al
    end;
   until (x1=x2);
  end
  else begin
   ai:=(dx-dy) shl 1;
   bi:=dx shl 1;
   d:=bi-dy;
   repeat
    if d>=0 then begin inc(x1,xi); inc(d,ai); end
    else inc(d,bi); inc(y1,yi);
    asm
     mov ax,$A000
     mov es,ax
     mov ax,_MaxX
     inc ax
     mul y1
     add ax,x1
     adc dx,0
     mov di,ax
     cmp dx,CurBank
     je @@2
     push dx
     call SwitchBank
@@2:
     mov al,color
     mov es:[di],al
    end;
   until (y1=y2);
  end;
end;

procedure _MaskedLine(x1,y1,x2,y2:word;color:byte;mask:word);
var d,dx,dy,ai,bi,xi,yi:integer;
begin
  if x1=x2 then begin
   if y1>y2 then begin d:=y2; y2:=y1; y1:=d; end;
   while y1<=y2 do begin
    asm
     rol mask,1
     jnc @@1
     mov ax,$A000
     mov es,ax
     mov ax,_MaxX
     inc ax
     mul y1
     add ax,x1
     adc dx,0
     mov di,ax
     cmp dx,CurBank
     je @@11
     push dx
     call SwitchBank
@@11:
     mov al,color
     mov es:[di],al
    @@1:
    end;
    inc(y1);
   end;
   Exit;
  end;
  if y1=y2 then begin
   if x1>x2 then begin d:=x2; x2:=x1; x1:=d; end;
   while x1<=x2 do begin
    asm
     rol mask,1
     jnc @@2
     mov ax,$A000
     mov es,ax
     mov ax,_MaxX
     inc ax
     mul y1
     add ax,x1
     adc dx,0
     mov di,ax
     cmp dx,CurBank
     je @@22
     push dx
     call SwitchBank
@@22:
     mov al,color
     mov es:[di],al
    @@2:
    end;
    inc(x1);
   end;
   Exit;
  end;
  if x1<x2 then begin xi:=1; dx:=x2-x1; end
  else begin xi:=-1; dx:=x1-x2; end;
  if y1<y2 then begin yi:=1; dy:=y2-y1; end
  else begin yi:=-1; dy:=y1-y2; end;
  asm
   rol mask,1
   jnc @@3
   mov ax,$A000
   mov es,ax
   mov ax,_MaxX
   inc ax
   mul y1
   add ax,x1
   adc dx,0
   mov di,ax
   cmp dx,CurBank
   je @@33
   push dx
   call SwitchBank
@@33:
   mov al,color
   mov es:[di],al
  @@3:
  end;
  if dx>dy then begin
   ai:=(dy-dx) shl 1;
   bi:=dy shl 1;
   d:=bi-dx;
   repeat
    if d>=0 then begin inc(y1,yi); inc(d,ai); end
    else inc(d,bi); inc(x1,xi);
    asm
     rol mask,1
     jnc @@4
     mov ax,$A000
     mov es,ax
     mov ax,_MaxX
     inc ax
     mul y1
     add ax,x1
     adc dx,0
     mov di,ax
     cmp dx,CurBank
     je @@44
     push dx
     call SwitchBank
@@44:
     mov al,color
     mov es:[di],al
    @@4:
    end;
   until (x1=x2);
  end
  else begin
   ai:=(dx-dy) shl 1;
   bi:=dx shl 1;
   d:=bi-dy;
   repeat
    if d>=0 then begin inc(x1,xi); inc(d,ai); end
    else inc(d,bi); inc(y1,yi);
    asm
     rol mask,1
     jnc @@5
     mov ax,$A000
     mov es,ax
     mov ax,_MaxX
     inc ax
     mul y1
     add ax,x1
     adc dx,0
     mov di,ax
     cmp dx,CurBank
     je @@55
     push dx
     call SwitchBank
@@55:
     mov al,color
     mov es:[di],al
    @@5:
    end;
   until (y1=y2);
  end;
end;

procedure _Box(x1,y1,x2,y2:word;color,fill:byte);assembler;
        (* This is pretty fast filled box *)
var bank,ofset,count,incr:word;
asm
 mov ax,_MaxX
 inc ax
 mov incr,ax
 cld

 push x1                        (* Draw left and right sides *)
 push y1
 push y2
 push word ptr color
 call _VLine
 push x2
 push y1
 push y2
 push word ptr color
 call _VLine
 push x1                        (* Draw top line of box *)
 push x2
 push y1
 push word ptr color
 call _HLine

 mov ax,$A000   (* Set up ES:0 for videobuffer segment *)
 mov es,ax
 mov ax,x2      (* Check if zero-width box *)
 sub ax,x1
 jz @@Exit
 dec ax
 jz @@Exit
 mov count,ax
 inc x1
 mov ax,y2      (* Check if zero-height box *)
 sub ax,y1
 jz @@Exit
 dec ax
 jz @@4
 inc y1

 mov al,_TransparentColor
 cmp al,fill                    (* filled rectangle? *)
 je @@4                         (* if no, go to bottom line drawing *)

 mov ax,_MaxX
 inc ax
 mul y1
 add ax,x1
 adc dx,0
 mov bank,dx
 mov ofset,ax
 cmp dx,CurBank
 je @@entry
 push bank
 mov cx,ax
 call SwitchBank
 mov ax,cx
 jmp @@entry

@@7:
 mov ax,ofset
 add ax,incr
 mov ofset,ax
 jnc @@3
 inc bank
 push bank
 call SwitchBank
@@3:

@@entry:
 add ax,count
 jc @@5                (* if next bank, goto @@5 *)
 mov di,ofset          (* if no, fill line *)
 mov cx,count
 mov al,fill
 mov ah,al
 mov bx,ax
 db $66; shl ax,16     (* shl eax,16 *)
 mov ax,bx
 mov bx,cx
 shr cx,2
 db 66h; rep stosw     (* rep stosd *)
 and bx,3
 mov cx,bx
 rep stosb
 jmp @@Exit2
@@5:
 mov di,ofset          (* fill part in current bank *)
 or cx,$FFFF           (* mov cx,$FFFF *)
 sub cx,di
 inc cx
 mov al,fill
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 mov bx,cx
 shr cx,2
 db 66h; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
 inc bank
 push bank             (* switch to next bank *)
 call SwitchBank
 dec bank
 xor di,di             (* and fill the rest of horizontal line *)
 mov cx,ofset
 add cx,count
 mov al,fill
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 mov bx,cx
 shr cx,2
 db 66h; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
@@Exit2:
 inc y1
 mov ax,y2
 cmp ax,y1              (* while y<y2 *)
 jne @@7

@@4:                    (* Draw bottom line *)
 dec x1
 push x1
 push x2
 push y2
 push word ptr color
 call _Hline
@@Exit:                 (* Voila! *)
end;

procedure _Fill(color:byte);assembler;
        (* Fill entire screen with color *)
var nbanks,bank,rest:word;
asm
 cld
 mov bx,_MaxX
 inc bx
 mov ax,_MaxY
 inc ax
 mul bx
 mov nbanks,dx
 mov rest,ax
 mov ax,$A000
 mov es,ax
 xor ax,ax
 mov bank,ax

@@1:
 push bank
 call SwitchBank
 xor di,di
 mov cx,8196
 mov al,color
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 db $66; rep stosw
 mov di,32768
 mov cx,8196
 mov al,color
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 db $66; rep stosw
 inc bank
 mov ax,bank
 dec ax
 cmp ax,nbanks
 jne @@1

 push bank
 call SwitchBank
 xor di,di
 mov cx,rest
 inc cx
 mov al,color
 mov ah,al
 mov bx,ax
 db $66; shl ax,16
 mov ax,bx
 mov bx,cx
 shr cx,2
 db $66; rep stosw
 and bx,3
 mov cx,bx
 rep stosb
end;

procedure _Ellipse(xc,yc,rx,ry:word;color:byte);
        (* When I look at this routine, my wish to
           rewrite it in assembler instantly disappears :) *)
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  _VLine(xc,yc-ry,yc+ry,color);
  Exit;
 end;
 if ry=0 then begin
  _HLine(xc-rx,xc+rx,yc,color);
  Exit;
 end;
 x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
 tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
 d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
 while dx<dy do begin
  _PutPixel(xc+x,yc+y,color);
  _PutPixel(xc-x,yc+y,color);
  _PutPixel(xc+x,yc-y,color);
  _PutPixel(xc-x,yc-y,color);
  if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
  Inc(x); dx:=dx+tbs; d:=d+bs+dx;
 end;
 d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
 while y>0 do begin
  _PutPixel(xc+x,yc+y,color);
  _PutPixel(xc-x,yc+y,color);
  _PutPixel(xc+x,yc-y,color);
  _PutPixel(xc-x,yc-y,color);
  if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
  y:=y-1; dy:=dy-tas; d:=d+as-dy;
 end;
 _PutPixel(xc+x,yc,color);
 _PutPixel(xc-x,yc,color);
end;

procedure _FilledEllipse(xc,yc,rx,ry:word;color:byte);
        (* Filled ellipse *)
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  _VLine(xc,yc-ry,yc+ry,color);
  Exit;
 end;
 if ry=0 then begin
  _HLine(xc-rx,xc+rx,yc,color);
  Exit;
 end;
 x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
 tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
 d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
 while dx<dy do begin
  _HLine(xc-x,xc+x,yc+y,color);
  _HLine(xc-x,xc+x,yc-y,color);
  if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
  Inc(x); dx:=dx+tbs; d:=d+bs+dx;
 end;
 d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
 while y>0 do begin
  _HLine(xc-x,xc+x,yc+y,color);
  _HLine(xc-x,xc+x,yc-y,color);
  if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
  y:=y-1; dy:=dy-tas; d:=d+as-dy;
 end;
 _HLine(xc-x,xc+x,yc,color);
end;

procedure _Triangle(x1,y1,x2,y2,x3,y3:word;color:byte);
begin
 _Line(x1,y1,x2,y2,color);
 _Line(x2,y2,x3,y3,color);
 _Line(x3,y3,x1,y1,color);
end;

procedure xchange(var a,b:integer);near;
var c:integer;
begin
 c:=a;a:=b;b:=c;
end;


procedure _FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
        (* This one is not too fast. What BP misses is 32-bit assembler... *)
var ymin,xmin,ymid,xmid,ymax,xmax,xmid2:longint;
    i:integer;
begin
 if y1>y2 then begin xchange(y1,y2); xchange(x1,x2); end;
 if y1>y3 then begin xchange(y1,y3); xchange(x1,x3); end;
 if y2>y3 then begin xchange(y2,y3); xchange(x2,x3); end;
 ymin:=y1; xmin:=x1; ymid:=y2; xmid:=x2; ymax:=y3; xmax:=x3;
 xmid2:=(xmax-xmin)*(ymid-ymin) div (ymax-ymin)+xmin;
 if xmid2<xmid then xchange(integer(xmid2),integer(xmid));
 if ymin<>ymid then
 for i:=ymin to ymid do
  _HLine((xmid-xmin)*(i-ymin) div (ymid-ymin)+xmin,
            (xmid2-xmin)*(i-ymin) div (ymid-ymin)+xmin,
            i,color);
 if ymid<>ymax then
 for i:=ymid to ymax do
  _HLine((xmax-xmid)*(i-ymid) div (ymax-ymid)+xmid,
            (xmax-xmid2)*(i-ymid) div (ymax-ymid)+xmid2,
            i,color);
end;

procedure _FilledPolygon(var VertexArray;NumOfVertexes,color:byte);
        (* Vertexes must be ordered, and polygon must be convex *)
type TVertex=record
      x,y:integer;
     end;
     TVertexArray=array[1..256] of TVertex;
begin
 if NumOfVertexes=1 then begin
  _PutPixel(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,color);
  Exit;
 end;
 if NumOfVertexes=2 then begin
  _Line(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
           TVertexArray(VertexArray)[2].x,TVertexArray(VertexArray)[2].y,
           color);
  Exit;
 end;
 repeat
  _FilledTriangle(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
               TVertexArray(VertexArray)[NumOfVertexes-1].x,TVertexArray(VertexArray)[NumOfVertexes-1].y,
               TVertexArray(VertexArray)[NumOfVertexes].x,TVertexArray(VertexArray)[NumOfVertexes].y,
               color);
  Dec(NumOfVertexes);
 until NumOfVertexes=2;
end;

(***************************************************************************

                         Drawing routines w/ clipping

****************************************************************************)

function OutCode(x,y:integer):integer;near;
var code:integer;
begin
 code:=0;
 if x<_MinClipX then code:=(code or 1);
 if y<_MinClipY then code:=(code or 2);
 if x>_MaxClipX then code:=(code or 4);
 if y>_MaxClipY then code:=(code or 8);
 OutCode:=code;
end;

function ClipLine(var x1,y1,x2,y2:integer):boolean;near;

(* This is Sutherland - Cohen clipping algorithm *)

var code1,code2:integer;
    inside,outside:boolean;
begin
 code1:=OutCode(x1,y1);
 code2:=OutCode(x2,y2);
 inside:=(code1 or code2)=0;
 outside:=(code1 and code2)<>0;
 while (not inside) and (not outside) do begin
  if code1=0 then begin
   xchange(x1,x2);
   xchange(y1,y2);
   xchange(code1,code2);
  end;
  if (code1 and 1)<>0 then begin
   y1:=y1+longint(y2-y1)*(_MinClipX-x1) div (x2-x1);
   x1:=_MinClipX;
  end else
  if (code1 and 2)<>0 then begin
   x1:=x1+longint(x2-x1)*(_MinClipY-y1) div (y2-y1);
   y1:=_MinClipY;
  end else
  if (code1 and 4)<>0 then begin
   y1:=y1+longint(y2-y1)*(_MaxClipX-x1) div (x2-x1);
   x1:=_MaxClipX;
  end else
  if (code1 and 8)<>0 then begin
   x1:=x1+longint(x2-x1)*(_MaxClipY-y1) div (y2-y1);
   y1:=_MaxClipY;
  end;
  code1:=OutCode(x1,y1);
  code2:=OutCode(x2,y2);
  inside:=(code1 or code2)=0;
  outside:=(code1 and code2)<>0;
 end;
 ClipLine:=(not outside);
end;

procedure __Line(x1,y1,x2,y2:integer;color:byte);
begin
 if ClipLine(x1,y1,x2,y2) then _Line(x1,y1,x2,y2,color);
end;

procedure __MaskedLine(x1,y1,x2,y2:integer;color:byte;Mask:word);
var MaskShift:word;
begin
 if x1<_MinClipX then MaskShift:=(_MinClipX-x1) mod 16 else MaskShift:=0;
 if ClipLine(x1,y1,x2,y2) then _MaskedLine(x1,y1,x2,y2,color,Mask shl MaskShift);
end;

procedure __HLine(x1,x2,y:integer;color:byte);
begin
 if (y<_MinClipY) or (y>_MaxClipY) or
    (x1>_MaxClipX) or (x2<_MinClipX) then Exit;
 if x1<_MinClipX then x1:=_MinClipX;
 if x2>_MaxClipX then x2:=_MaxClipX;
 _HLine(x1,x2,y,color);
end;

procedure __VLine(x,y1,y2:integer;color:byte);
begin
 if (x<_MinClipX) or (x>_MaxClipX) or
    (y1>_MaxClipY) or (y2<_MinClipY) then Exit;
 if y1<_MinClipY then y1:=_MinClipY;
 if y2>_MaxClipY then y2:=_MaxClipY;
 _VLine(x,y1,y2,color);
end;

procedure __Box(x1,y1,x2,y2:integer;color,fill:byte);
begin
 __HLine(x1,x2,y1,color);
 __VLine(x1,y1,y2,color);
 __VLine(x2,y1,y2,color);
 __HLine(x1,x2,y2,color);
 if fill<>_TransparentColor then begin
  if (x1>_MaxClipX) or (y1>_MaxClipY) or
     (x2<_MinClipX) or (y2<_MinClipY) then Exit;
  if x1<_MinClipX then x1:=_MinClipX;
  if y1<_MinClipY then y1:=_MinClipY;
  if x2>_MaxClipX then x2:=_MaxClipX;
  if y2>_MaxClipY then Y2:=_MaxClipY;
  _Box(x1+1,y1+1,x2-1,y2-1,fill,fill);
 end;
end;

procedure __Ellipse(xc,yc:integer;rx,ry:word;color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  __VLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  __HLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
 tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
 d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
 while dx<dy do begin
  __PutPixel(xc+x,yc+y,Color);
  __PutPixel(xc-x,yc+y,Color);
  __PutPixel(xc+x,yc-y,Color);
  __PutPixel(xc-x,yc-y,Color);
  if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
  Inc(x); dx:=dx+tbs; d:=d+bs+dx;
 end;
 d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
 while y>0 do begin
  __PutPixel(xc+x,yc+y,Color);
  __PutPixel(xc-x,yc+y,Color);
  __PutPixel(xc+x,yc-y,Color);
  __PutPixel(xc-x,yc-y,Color);
  if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
  y:=y-1; dy:=dy-tas; d:=d+as-dy;
 end;
 __PutPixel(xc+x,yc,Color);
 __PutPixel(xc-x,yc,Color);
end;

procedure __FilledEllipse(xc,yc:integer;rx,ry:word;Color:byte);
var x,y:word;
    a,b,as,tas,bs,tbs:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  __VLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  __HLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx; b:=ry; as:=a*a;
 tas:=as shl 1; bs:=b*b; tbs:=bs shl 1;
 d:=bs-as*b+(as shr 2); dx:=0; dy:=tas*b;
 while dx<dy do begin
  __HLine(xc-x,xc+x,yc+y,Color);
  __HLine(xc-x,xc+x,yc-y,Color);
  if d>0 then begin y:=y-1; dy:=dy-tas; d:=d-dy; end;
  Inc(x); dx:=dx+tbs; d:=d+bs+dx;
 end;
 d:=d+((3*(as-bs)div 2-(dx+dy))div 2);
 while y>0 do begin
  __HLine(xc-x,xc+x,yc+y,Color);
  __HLine(xc-x,xc+x,yc-y,Color);
  if d<0 then begin Inc(x); dx:=dx+tbs; d:=d+dx; end;
  y:=y-1; dy:=dy-tas; d:=d+as-dy;
 end;
 __HLine(xc-x,xc+x,yc,Color);
end;

procedure __Triangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
begin
 __Line(x1,y1,x2,y2,color);
 __Line(x2,y2,x3,y3,color);
 __Line(x3,y3,x1,y1,color);
end;

procedure __FilledTriangle(x1,y1,x2,y2,x3,y3:integer;color:byte);
var ymin,xmin,ymid,xmid,ymax,xmax,xmid2:longint;
    i:integer;
begin
 if y1>y2 then begin xchange(y1,y2); xchange(x1,x2); end;
 if y1>y3 then begin xchange(y1,y3); xchange(x1,x3); end;
 if y2>y3 then begin xchange(y2,y3); xchange(x2,x3); end;
 ymin:=y1; xmin:=x1; ymid:=y2; xmid:=x2; ymax:=y3; xmax:=x3;
 xmid2:=(xmax-xmin)*(ymid-ymin) div (ymax-ymin)+xmin;
 if xmid2<xmid then xchange(integer(xmid2),integer(xmid));
 if ymin<>ymid then
 for i:=ymin to ymid do
  __HLine((xmid-xmin)*(i-ymin) div (ymid-ymin)+xmin,
            (xmid2-xmin)*(i-ymin) div (ymid-ymin)+xmin,
            i,color);
 if ymid<>ymax then
 for i:=ymid to ymax do
  __HLine((xmax-xmid)*(i-ymid) div (ymax-ymid)+xmid,
            (xmax-xmid2)*(i-ymid) div (ymax-ymid)+xmid2,
            i,color);
end;

procedure __FilledPolygon(var VertexArray;NumOfVertexes,color:byte);
        (* Vertexes must be ordered, and polygon must be convex *)
type TVertex=record
      x,y:integer;
     end;
     TVertexArray=array[1..256] of TVertex;
begin
 if NumOfVertexes=1 then begin
  __PutPixel(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,color);
  Exit;
 end;
 if NumOfVertexes=2 then begin
  __Line(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
           TVertexArray(VertexArray)[2].x,TVertexArray(VertexArray)[2].y,
           color);
  Exit;
 end;
 repeat
  __FilledTriangle(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
               TVertexArray(VertexArray)[NumOfVertexes-1].x,TVertexArray(VertexArray)[NumOfVertexes-1].y,
               TVertexArray(VertexArray)[NumOfVertexes].x,TVertexArray(VertexArray)[NumOfVertexes].y,
               color);
  Dec(NumOfVertexes);
 until NumOfVertexes=2;
end;

(***************************************************************************

                        Image manipulating procedures

****************************************************************************)


procedure XOR16(var Source,Dest;Count:word);near;assembler;
asm
 push ds
 les di,Dest
 lds si,Source
 mov cx,Count
 mov bx,cx
 shr cx,1
@@1:
 lodsw
 xor word ptr es:[di],ax
 inc di
 loop @@1
 and bx,1
 jz @@exit
 lodsb
 xor byte ptr es:[di],al
@@exit:
 pop ds
end;

procedure _GetImage(x1,y1,x2,y2:word;clip:pointer);
        (* Copies image part to buffer *)
var SourcPtr,DestPtr:pointer;
    bank1,ofset1,bank2,ofset2,i:word;
    count,DestOfset:word;
begin
 DestOfset:=Ofs(clip^);
 for i:=y1 to y2 do begin
  asm
   mov ax,_MaxX
   inc ax
   mul i
   mov bx,ax
   mov cx,dx
   add ax,x1
   adc dx,0
   mov bank1,dx
   mov ofset1,ax
   add bx,x2
   adc cx,0
   mov bank2,cx
   mov ofset2,bx
  end;
  if bank1<>CurBank then SwitchBank(bank1);
   SourcPtr:=Ptr($A000,ofset1);
   DestPtr:=Ptr(Seg(clip^),DestOfset);
  if bank1=bank2 then begin
   count:=x2-x1+1;
   Move(SourcPtr^,DestPtr^,count);
   inc(DestOfset,count);
  end
  else begin
   count:=65535-ofset1+1;
   Move(SourcPtr^,DestPtr^,count);
   inc(DestOfset,count);
   DestPtr:=Ptr(Seg(clip^),DestOfset);
   SwitchBank(bank2);
   SourcPtr:=Ptr($A000,0);
   Move(SourcPtr^,DestPtr^,ofset2+1);
   inc(DestOfset,ofset2+1);
  end;
 end;
end;

procedure _PutImage(x1,y1:integer;width,height:word;clip:pointer);
        (* Puts previously copied image *)
var SourcPtr,DestPtr:pointer;
    bank1,ofset1,bank2,ofset2,i,x2:word;
    SourcOfset,count,xshift,yshift:word;
begin
 if (x1>_MaxX) or (y1>_MaxY) then Exit;
 xshift:=0;yshift:=0;
 if x1<0 then begin
  if x1+width<=0 then Exit;
  xshift:=-x1;
  width:=width+x1;
  x1:=0;
 end;
 if y1<0 then begin
  if y1+height<=0 then Exit;
  yshift:=-y1;
  height:=height+y1;
  y1:=0;
 end;
 SourcOfset:=ofs(clip^)+yshift*(width+xshift)+xshift;
 x2:=x1+width-1;
 if x2>_MaxX then begin
  inc(xshift,x2-_MaxX);
  width:=_MaxX-x1+1;
  x2:=_MaxX;
 end;
 if (y1+height-1)>_MaxY then height:=_MaxY-y1+1;
 for i:=y1 to y1+height-1 do begin
  asm
   mov ax,_MaxX
   inc ax
   mul i
   mov bx,ax
   mov cx,dx
   add ax,x1
   adc dx,0
   mov bank1,dx
   mov ofset1,ax
   add bx,x2
   adc cx,0
   mov bank2,cx
   mov ofset2,bx
  end;
  SourcPtr:=ptr(seg(clip^),SourcOfset);
  DestPtr:=Ptr($A000,ofset1);
  if bank1<>CurBank then SwitchBank(bank1);
  if bank1=bank2 then begin
   Move(SourcPtr^,DestPtr^,width);
   inc(SourcOfset,width+xshift);
  end
  else begin
   count:=65535-ofset1+1;
   Move(SourcPtr^,DestPtr^,count);
   inc(SourcOfset,count);
   SourcPtr:=Ptr(Seg(clip^),SourcOfset);
   SwitchBank(bank2);
   DestPtr:=Ptr($A000,0);
   Move(SourcPtr^,DestPtr^,ofset2+1);
   inc(SourcOfset,ofset2+1+xshift);
  end;
 end;
end;

procedure _XORImage(x1,y1:integer;width,height:word;clip:pointer);
var SourcPtr,DestPtr:pointer;
    bank1,ofset1,bank2,ofset2,i,x2:word;
    SourcOfset,count,xshift,yshift:word;
begin
 if (x1>_MaxX) or (y1>_MaxY) then Exit;
 xshift:=0;yshift:=0;
 if x1<0 then begin
  if x1+width<=0 then Exit;
  xshift:=-x1;
  width:=width+x1;
  x1:=0;
 end;
 if y1<0 then begin
  if y1+height<=0 then Exit;
  yshift:=-y1;
  height:=height+y1;
  y1:=0;
 end;
 SourcOfset:=ofs(clip^)+yshift*(width+xshift)+xshift;
 x2:=x1+width-1;
 if x2>_MaxX then begin
  inc(xshift,x2-_MaxX);
  width:=_MaxX-x1+1;
  x2:=_MaxX;
 end;
 if (y1+height-1)>_MaxY then height:=_MaxY-y1+1;
 for i:=y1 to y1+height-1 do begin
  asm
   mov ax,_MaxX
   inc ax
   mul i
   mov bx,ax
   mov cx,dx
   add ax,x1
   adc dx,0
   mov bank1,dx
   mov ofset1,ax
   add bx,x2
   adc cx,0
   mov bank2,cx
   mov ofset2,bx
  end;
  SourcPtr:=ptr(seg(clip^),SourcOfset);
  DestPtr:=Ptr($A000,ofset1);
  if bank1<>CurBank then SwitchBank(bank1);
  if bank1=bank2 then begin
   XOR16(SourcPtr^,DestPtr^,width);
   inc(SourcOfset,width);
   inc(SourcOfset,xshift);
  end
  else begin
   count:=65535-ofset1+1;
   XOR16(SourcPtr^,DestPtr^,count);
   inc(SourcOfset,count);
   SourcPtr:=Ptr(Seg(clip^),SourcOfset);
   SwitchBank(bank2);
   DestPtr:=Ptr($A000,0);
   XOR16(SourcPtr^,DestPtr^,ofset2+1);
   inc(SourcOfset,ofset2+1);
   inc(SourcOfset,xshift);
  end;
 end;
end;

procedure _PutTransparentImage(x1,y1:integer;width,height:word;clip:pointer);
        (* Puts everything except transparent pixels.
           This one is slow. No assembler at all... Ick.. *)
var i,j,SourcOfset:word;
    xshift,yshift:word;
begin
 if (x1>_MaxX) or (y1>_MaxY) then Exit;
 xshift:=0;yshift:=0;
 if x1<0 then begin
  if x1+width<=0 then Exit;
  xshift:=-x1;
  width:=width+x1;
  x1:=0;
 end;
 if y1<0 then begin
  if y1+height<=0 then Exit;
  yshift:=-y1;
  height:=height+y1;
  y1:=0;
 end;
 SourcOfset:=ofs(clip^)+yshift*(width+xshift)+xshift;
 if (x1+width-1)>_MaxX then begin
  inc(xshift,x1+width-1-_MaxX);
  width:=_MaxX-x1+1;
 end;
 if (y1+height-1)>_MaxY then height:=_MaxY-y1+1;
 for i:=y1 to y1+height-1 do begin
  for j:=x1 to x1+width-1 do
   _PutTransparentPixel(j,i,Mem[seg(clip^):SourcOfset+j-x1]);
  inc(SourcOfset,width+xshift);
 end;
end;

procedure _GetLargeImage(x1,y1,x2,y2:word;handle:word);
        (* Gets image of size > 64 Kb and stores it in XMS *)
var blocks,tail,height,i,size:word;
    y:integer;
    p:pointer;
begin
 height:=64000 div (x2-x1+1);
 size:=height*(x2-x1+1);
 blocks:=(y2-y1+1) div height;
 tail:=(y2-y1+1) mod height;
 y:=y1;
 GetMem(p,64000);
 with XMDescriptor do begin
  Length:=size;
  if Odd(Length) then Inc(Length);
  SourceHandle:=0;
  SourceOffset:=seg(p^)*65536+ofs(p^);
  DestHandle:=handle;
  DestOffset:=0;
 end;
 for i:=1 to blocks do begin
  _GetImage(x1,y,x2,y+height-1,p);
  xmmMoveEMB(XMDescriptor);
  inc(XMDescriptor.DestOffset,size);
  inc(y,height);
 end;
 XMDescriptor.Length:=tail*(x2-x1+1);
 if odd(XMDescriptor.Length) then inc(XMDescriptor.Length);
 _GetImage(x1,y,x2,y+tail-1,p);
 xmmMoveEMB(XMDescriptor);
 FreeMem(p,64000);
end;

procedure _PutLargeImage(x1,y1:integer;width,height:word;handle:word);
        (* Puts large image from XMS buffer *)
var blocks,tail,h,i,size:word;
    y:integer;
    p:pointer;
begin
 h:=64000 div width;
 size:=h*width;
 blocks:=height div h;
 tail:=height mod h;
 y:=y1;
 GetMem(p,64000);
 with XMDescriptor do begin
  Length:=size;
  if Odd(Length) then Inc(Length);
  SourceHandle:=handle;
  SourceOffset:=0;
  DestHandle:=0;
  DestOffset:=seg(p^)*65536+ofs(p^);
 end;
 for i:=1 to blocks do begin
  if y>_MaxY then begin
   FreeMem(p,64000);
   Exit;
  end;
  if y>=-height then begin
   xmmMoveEMB(XMDescriptor);
   _PutImage(x1,y,width,h,p);
  end;
  inc(XMDescriptor.SourceOffset,size);
  inc(y,h);
 end;
 XMDescriptor.Length:=tail*width;
 if odd(XMDescriptor.Length) then inc(XMDescriptor.Length);
 xmmMoveEMB(XMDescriptor);
 _PutImage(x1,y,width,tail,p);
 FreeMem(p,64000);
end;

(***************************************************************************

                        Image files manipulating routines

****************************************************************************)

{$I PCXHEADR.INC}

procedure _LoadPCX(FileName:string;FileOffset:longint;x,y:word);
var InCounter,OutCounter:word;
    i,LineCounter:word;
    Header:T_PCX_Header;
    Dummy:word;
    Count,Data:byte;
    f:file;
    _IO_Buffer:array[1..8192] of byte;
begin
 Assign(f,FileName);
 Reset(f,1);
 Seek(f,FileOffset);
 BlockRead(f,Header,128);
 with Header do
  if (Signature<>$0A) or (BPP<>8) or (Planes<>1) or
     ((XMax-XMin+1)>4096) then begin
   Close(f);
   Exit;
  end;
 InCounter:=1;OutCounter:=1;LineCounter:=0;
 BlockRead(f,_IO_Buffer,4096,Dummy);
 repeat
  if (_IO_Buffer[InCounter] and $C0)=$C0 then begin
   Count:=(_IO_Buffer[InCounter]) and $3F;
   Inc(InCounter);
   if InCounter=4097 then begin
    InCounter:=1;
    BlockRead(f,_IO_Buffer,4096,Dummy);
   end;
   Data:=_IO_Buffer[InCounter];
   Inc(InCounter);
   if InCounter=4097 then begin
    InCounter:=1;
    BlockRead(f,_IO_Buffer,4096,Dummy);
   end;
  end
  else begin
   Count:=1;
   Data:=_IO_Buffer[InCounter];
   Inc(InCounter);
   if InCounter=4097 then begin
    InCounter:=1;
    BlockRead(f,_IO_Buffer,4096,Dummy);
   end;
  end;
  for i:=1 to Count do begin
   _IO_Buffer[OutCounter+4096]:=Data;
   Inc(OutCounter);
   if OutCounter=Header.BPL+1 then begin
    OutCounter:=1;
    Inc(LineCounter);
    _PutImage(x,y,Header.XMax-Header.XMin+1,1,@_IO_Buffer[4097]);
    Inc(y);
   end;
  end;
 until LineCounter=Header.YMax-Header.YMin+1;
 Close(f);
end;

procedure _GetPCXSize(filename:string;fileoffset:longint;var x,y:word);
var Header:T_PCX_Header;
    f:file;
begin
 x:=0; y:=0;
 Assign(f,FileName);
 Reset(f,1);
 Seek(f,FileOffset);
 BlockRead(f,Header,128);
 with Header do
  if (Signature<>$0A) or (BPP<>8) or (Planes<>1) then begin
   Close(f);
   Exit;
  end;
 x:=Header.XMax-Header.XMin+1;
 y:=Header.YMax-Header.YMin+1;
 Close(f);
end;

procedure _SavePCX(FileName:string;x1,y1,x2,y2:word);
var Header:T_PCX_Header;
    Palette256:array[-1..767] of byte;
    f:file;
    Len,i,InPos,OutPos:word;
    Data,Count:byte;
    _IO_Buffer:array[1..8192] of byte;
begin
 Len:=x2-x1+1;
 if Len>4096 then Exit;
 _GetPalette(Palette256[0]);
 for i:=0 to 767 do Palette256[i]:=round(Palette256[i]*255/63);
 Palette256[-1]:=$0C;
 with Header do begin
  Signature:=$0A;
  Version:=5;
  Encoding:=1;
  BPP:=8;
  XMin:=x1;
  YMin:=y1;
  XMax:=x2;
  YMax:=y2;
  PPIx:=Round((_MaxX+1)/11);
  PPIy:=Round((_MaxY+1)/8.5);
  for i:=0 to 47 do Palette[i]:=Palette256[i];
  Reserved1:=0;
  Planes:=1;
  BPL:=Len; if Odd(BPL) then Inc(BPL);
  PaletteDescriptor:=1;
  XResolution:=_MaxX+1;
  YResolution:=_MaxY+1;
  for i:=0 to 53 do Reserved2[i]:=0;
 end;
 Assign(f,FileName);
 Rewrite(f,1);
 BlockWrite(f,Header,128);
 OutPos:=1;
 for i:=y1 to y2 do begin
  _GetImage(x1,i,x1+Len-1,i,@_IO_Buffer);
  InPos:=1;
  repeat
   Data:=_IO_Buffer[InPos];Count:=1;
   Inc(InPos);
   repeat
    if InPos=Header.BPL+1 then break;
    if _IO_Buffer[InPos]=Data then begin
     Inc(Count);
     Inc(InPos);
    end
    else break;
   until Count=63;
   if (Count>1) or ((Data and $C0)=$C0) then begin
    _IO_Buffer[OutPos+4096]:=($C0 or Count); Inc(OutPos);
    if OutPos=4097 then begin OutPos:=1; BlockWrite(f,_IO_Buffer[4097],4096); end;
   end;
   _IO_Buffer[OutPos+4096]:=Data; Inc(OutPos);
   if OutPos=4097 then begin OutPos:=1; BlockWrite(f,_IO_Buffer[4097],4096); end;
   if InPos=Header.BPL+1 then break;
  until false;
 end;
 if OutPos>1 then BlockWrite(f,_IO_Buffer[4097],OutPos-1);
 BlockWrite(f,Palette256[-1],769);
 Close(f);
end;

procedure _GetPCXPalette(FileName:string;var Palette);
var f:file;
    Test:byte;
    i:word;
type TPalette=array[0..767] of byte;
begin
 Assign(f,FileName);
 Reset(f,1);
 Seek(f,FileSize(f)-769);
 BlockRead(f,Test,1);
 if Test<>$0C then begin
  Close(f);
  Exit;
 end;
 BlockRead(f,Palette,768);
 for i:=0 to 767 do TPalette(Palette)[i]:=round(TPalette(Palette)[i]*63/255);
 Close(f);
end;

(***************************************************************************

                        Fonts and text strings handling

****************************************************************************)

type TFont=array[0..255,0..15] of byte;

procedure _WriteStr(x,y:word;S:string;color:byte);
var j:word;
    i,k:byte;
begin
 if FontPtr=nil then Exit;
 for i:=1 to Length(S) do begin
  for j:=0 to 15 do begin
   k:=TFont(FontPtr^)[Ord(S[i]),j];
   asm
    mov cx,8
@@2:
    rol k,1
    jnc @@1
    push cx
    mov bx,x
    add bx,8
    sub bx,cx
    push bx
    mov bx,y
    add bx,j
    push bx
    push word ptr color
    call _PutPixel
    pop cx
@@1:
    loop @@2
   end;
  end;
  inc(x,8);
 end;
end;

procedure _SetFont(p:pointer);
begin
  FontPtr:=p;
end;

(***************************************************************************

                                Miscellaneous

****************************************************************************)

procedure _WaitVSync;assembler;
        (* Waits for vertical retrace start *)
asm
 mov dx,$3DA
@@0:
 in al,dx
 test al,8
 jnz @@0
@@1:
 in al,dx
 test al,8
 jz @@1
end;

procedure _WaitVSyncStart;assembler;
asm
 mov dx,$3DA
@@1:
 in al,dx
 test al,8
 jz @@1
end;

procedure _SetTextMode;
        (* Set 80x25 color text *)
begin
 if @_ShutDown<>nil then _ShutDown;
 asm
  mov ax,3
  int $10
 end;
 _Mode:=0;
end;

function Copyleft_Id:string;
begin
 Copyleft_Id:=Copyleft;
end;

procedure _RestoreDisplayBounds;
begin
 _MinClipX:=0;
 _MinClipY:=0;
 _MaxClipX:=_MaxX;
 _MaxClipY:=_MaxY;
end;

(****************************************************************************

                        Virtual display control

****************************************************************************)

procedure _SetLogicalWidth(NewWidth:word);assembler;
asm
 mov ax,$4F06
 xor bl,bl
 mov cx,NewWidth
 int $10
 cmp ax,$004F
 jne @@Exit
 dec cx
 mov _MaxX,cx
 dec dx
 mov _MaxY,dx
 @@Exit:
end;

procedure _SetDisplayOrigin(x,y:word);assembler;
asm
 mov ax,$4F07
 mov bx,$0080
 mov cx,x
 mov dx,y
 int $10
end;

(***************************************************************************

                        Initialization

****************************************************************************)

begin
 FindVESA;
 if _VESAPresent then _Init(_VESA);
 Copyleft:=#13+#10+' SuperVGA library Version '+Version+'.DOS'+#13+#10+
           ' Virtual Research independent group production (C) 1996-1999'+#13+#10;
end.
