(*************************************************)
(* Extended memory manager API for XMS 2.0       *)
(* Unit version: 2.0 / 02.Jan.1998               *)
(* (C)oded by Asp, 1998                          *)
(* Virtual Research independent group production *)
(*************************************************)

unit xmanager;
{$S-,R-,Q-,V-,B-,X+,G+,F+}

interface

var xmmPresent:boolean; (* Is the Xtended Memory Manager installed? *)
    xmmXMSVersion:word;         (* XMS version *)
    xmmDriverVersion:word;      (* XMM version *)
    xmmHMAPresent:boolean;      (* Is the HMA exist? *)
    xmmResult:byte;             (* Error code *)

type TxmmMoveStruc=record
      Length:longint;           (* Number of bytes to transfer *)
      SourceHandle:word;        (* Handle of source block *)
      SourceOffset:longint;     (* Offset into source *)
      DestHandle:word;          (* Handle of destination block *)
      DestOffset:longint;       (* Offset into destination block *)
     end;

(* If SourceHandle is set to $0000, the SourceOffset is interpreted
   as a standard segment:offset pair which refers to memory that is
   directly accessible by the processor.  The segment:offset pair
   is stored in Intel DWORD notation (offset in lower word).
   The same is true for DestHandle and DestOffset. Length must be even. *)

function xmmRequestHMA(Size:word):boolean;
function xmmReleaseHMA:boolean;
function xmmGlobalEnableA20:boolean;
function xmmGlobalDisableA20:boolean;
function xmmLocalEnableA20:boolean;
function xmmLocalDisableA20:boolean;
function xmmQueryA20:boolean;
function xmmMemAvail:word;
function xmmMaxAvail:word;
function xmmAllocEMB(Size:word;var Handle:word):boolean;
function xmmFreeEMB(Handle:word):boolean;
function xmmMoveEMB(var Descriptor:TxmmMoveStruc):boolean;
function xmmLockEMB(Handle:word;var Address:longint):boolean;
function xmmUnlockEMB(Handle:word):boolean;
function xmmGetEMBInfo(Handle:word;var LockCount,FreeEMBs:byte;var Size:word):boolean;
function xmmReallocEMB(Handle,NewSize:word):boolean;
function xmmRequestUMB(Size:word;var Segment,ActualSize:word):boolean;
function xmmReleaseUMB(Segment:word):boolean;


implementation

var xmmAPIPtr:longint;  (* Pointer to XMM API interface *)

procedure FindXMM;assembler;
        (* Is an XMS driver installed? *)
asm
 mov ax,$4300
 int $2F
 cmp al,$80
 jne @NotFound
 mov xmmPresent,true
 mov ax,$4310
 int $2F
 mov word ptr [xmmAPIPtr],bx
 mov word ptr [xmmAPIPtr+2],es
 xor ah,ah
 call [xmmAPIPtr]
 mov xmmXMSVersion,ax
 mov xmmDriverVersion,bx
 cmp dx,0
 je @NoHMA
 mov xmmHMAPresent,true
 jmp @exit
@NoHMA:
 mov xmmHMAPresent,false
 jmp @exit
@NotFound:
 mov xmmPresent,false
@exit:
end;

(***************************************************************************

                        The HMA management routines

****************************************************************************)

function xmmRequestHMA(Size:word):boolean;assembler;
        (* This function attempts to reserve the 64K-16
           byte high memory area for the caller.
           If the caller is a TSR or device driver,
           Size = Space needed in the HMA by the caller in bytes,
           if the caller is an application program, Size should be
           equal to $FFFF. Function returns True on success,
           otherwise returns False. *)
asm
 mov ah,1
 mov dx,Size
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmReleaseHMA:boolean;assembler;
        (* This function releases the high memory area
           and allows other programs to use it. Programs
           which allocate the HMA must release it before exiting. *)
asm
 mov ah,2
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

(***************************************************************************

                        The A20 gate management routines

****************************************************************************)

function xmmGlobalEnableA20:boolean;assembler;
        (* This function attempts to enable the A20 line.
           It should only be used by programs which have control of the HMA.
           The A20 line should be turned off before a program releases
           control of the system. *)
asm
 mov ah,3
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmGlobalDisableA20:boolean;assembler;
        (* This function attempts to disable the A20 line.
           It should only be used by programs which have control of the HMA.
           The A20 line should be disabled before a program releases control
           of the system. *)
asm
 mov ah,4
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmLocalEnableA20:boolean;assembler;
        (* This function attempts to enable the A20 line.
           It should only be used by programs which need direct access
           to extended memory.  Programs which use this function should
           call Function 6 (Local Disable A20) before releasing
           control of the system. *)
asm
 mov ah,5
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmLocalDisableA20:boolean;assembler;
        (* This function cancels a previous call to Function 5
           (Local Enable A20). It should only be used by programs
           which need direct access to extended memory. Previous calls
           to Function 5 must be canceled before releasing control
           of the system. *)
asm
 mov ah,6
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmQueryA20:boolean;assembler;
        (* This function checks to see if the A20 line
           is physically enabled. *)
asm
 mov ah,7
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

(***************************************************************************

                The eXtended Memory management routines

****************************************************************************)

function xmmMemAvail:word;assembler;
        (* Returns total amount of free extended memory
           in Kbytes *)
asm
 mov ah,8
 xor bl,bl
 call [xmmAPIPtr]
 mov ax,dx
 mov xmmResult,bl
end;

function xmmMaxAvail:word;assembler;
        (* Returns size of the largest free EMB in Kbytes *)
asm
 mov ah,8
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmAllocEMB(Size:word;var Handle:word):boolean;assembler;
        (* This function attempts to allocate a block
           of the given size out of the pool of free extended memory. *)
asm
 mov ah,9
 mov dx,Size
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
 les di,Handle
 mov word ptr [es:di],dx
end;

function xmmFreeEMB(Handle:word):boolean;assembler;
        (* This function frees a block of extended memory
           which was previously allocated using Function 9.
           Programs which allocate extended memory should free their
           memory blocks before exiting. *)
asm
 mov ah,$A
 mov dx,Handle
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmMoveEMB(var Descriptor:TxmmMoveStruc):boolean;assembler;
        (* This function attempts to transfer a block of data
           from one location to another. *)
asm
 push ds
 mov ah,$B
 xor bl,bl
 lds si,Descriptor
 call [xmmAPIPtr]
 mov xmmResult,bl
 pop ds
end;

function xmmLockEMB(Handle:word;var Address:longint):boolean;assembler;
        (* This function locks an extended memory block and
           returns its base address as a 32-bit linear address.
           Locked memory blocks are guaranteed not to move.
           The 32-bit pointer is only valid while the block is locked.
           Locked blocks should be unlocked as soon as possible. *)
asm
 mov ah,$C
 mov dx,Handle
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
 les di,Address
 mov word ptr [es:di],bx
 mov word ptr [es:di+2],dx
end;

function xmmUnlockEMB(Handle:word):boolean;assembler;
        (* This function unlocks a locked extended memory block.
           Any 32-bit pointers into the block become invalid
           and should no longer be used. *)
asm
 mov ah,$D
 mov dx,handle
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

function xmmGetEMBInfo(Handle:word;var LockCount,FreeEMBs:byte;var Size:word):boolean;assembler;
        (* This function returns additional information about
           an extended memory block to the caller. *)
asm
 mov ah,$E
 mov dx,Handle
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
 les di,LockCount
 mov byte ptr [es:di],bh
 les di,FreeEMBs
 mov byte ptr [es:di],bl
 les di,Size
 mov [es:di],dx
end;

function xmmReallocEMB(Handle,NewSize:word):boolean;assembler;
asm
 mov ah,$F
 mov bx,NewSize
 mov dx,Handle
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

(***************************************************************************

                The Upper Memory Blocks management routines

****************************************************************************)

function xmmRequestUMB(Size:word;var Segment,ActualSize:word):boolean;assembler;
        (* Size is required size in PARAGRAPHS (16-byte units) *)
asm
 mov ah,$10
 mov dx,Size
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
 les di,Segment
 mov word ptr [es:di],bx
 les di,ActualSize
 mov word ptr [es:di],dx
end;

function xmmReleaseUMB(Segment:word):boolean;assembler;
asm
 mov ah,$11
 mov dx,Segment
 xor bl,bl
 call [xmmAPIPtr]
 mov xmmResult,bl
end;

(***************************************************************************

                                Initialization

****************************************************************************)

begin
 FindXMM;
end.

(***************************************************************************

                        Appendix: XMS error codes.

****************************************************************************

 The xmmResult variable equals to BL on function's exit. In most cases
 xmmResult=0 means no error. After call to xmmGetEMBInfo xmmResult
 contains number of free EMB handles. After call to xmmRequestUMB
 xmmResult contains lower byte of the allocated UMB segment. In all
 other cases xmmResult contains error code:

       80h if the function is not implemented
       81h if a VDISK device is detected
       82h if an A20 error occurs
       8Eh if a general driver error occurs
       8Fh if an unrecoverable driver error occurs
       90h if the HMA does not exist
       91h if the HMA is already in use
       92h if DX is less than the /HMAMIN= parameter
       93h if the HMA is not allocated
       94h if the A20 line is still enabled
       A0h if all extended memory is allocated
       A1h if all available extended memory handles are in use
       A2h if the handle is invalid
       A3h if the SourceHandle is invalid
       A4h if the SourceOffset is invalid
       A5h if the DestHandle is invalid
       A6h if the DestOffset is invalid
       A7h if the Length is invalid
       A8h if the move has an invalid overlap
       A9h if a parity error occurs
       AAh if the block is not locked
       ABh if the block is locked
       ACh if the block's lock count overflows
       ADh if the lock fails
       B0h if a smaller UMB is available
       B1h if no UMBs are available
       B2h if the UMB segment number is invalid

*)
