(***************************************************************************)
(* Miscellaneous bitmap effects for 256-color bitmaps (or is it better to  *)
(* say "bytemaps"? :)                                                      *)
(* Coded by Asp //Virtual Research                                         *)
(* Unit version : 1.0                                                      *)
(* Last updated : 05.Jul.1999                                              *)
(*                                                                         *)
(* Rotation procedures are portionally based on sources by                 *)
(* Lewis A. Sellers //The Minimalist Group                                 *)
(***************************************************************************)

(*
   Note: This unit is in no way optimized. The purpose was to write
   _universal_ routines, able to work with bitmaps of (almost) any
   size and with (almost) any input parameters. There are obvious
   ways to dramatically improve performance of some routines, like:
   - port them to inline assembly;
   - rewrite them for fixed-size bitmaps;
   - rewrite them for fixed screen size;
   - use precalculated scale factors, pixel offsets etc.

   However, all of procedures were tested and they seem to do
   the job. They are still pretty fast, at least on Pentium machine.

   This code is released to public domain. Enjoy.

      - Asp / Virtual Research independent group coder.
*)


unit bmfx;

interface

type TBitmap=record
 Width:word;
 Height:word;
 case byte of
  0:(Body:pointer);
  1:(_Body:dword);
end;

procedure bmRescaleX(OldBitmap,NewBitmap:TBitmap);
procedure bmRescaleY(OldBitmap,NewBitmap:TBitmap);
procedure bmRescaleXY(OldBitmap,NewBitmap:TBitmap);
procedure bmShiftLine(Bitmap:TBitmap;Line,ShiftValue:word);
procedure bmShiftLineZ(Bitmap:TBitmap;Line,ShiftValue:word;FillColor:byte);
procedure bmShiftColumn(Bitmap:TBitmap;Column,ShiftValue:word);
procedure bmShiftColumnZ(Bitmap:TBitmap;Column,ShiftValue:word;FillColor:byte);
procedure bmRotate(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle:word);
procedure bmRotateZ(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle:word;FillColor:byte);
procedure bmRotScale(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle,ScaleFactor:word);
procedure bmRotScaleZ(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle,ScaleFactor:word;FillColor:byte);
procedure bmHalf(OldBitmap,NewBitmap:TBitmap);
procedure bmDouble(OldBitmap,NewBitmap:TBitmap);
procedure bmHalfSmooth(OldBitmap,NewBitmap:TBitmap;TransTable:pointer);
procedure bmDoubleInterlace(OldBitmap,NewBitmap:TBitmap);
procedure bmSmooth(Bitmap:TBitmap;TransTable:pointer);
procedure bmAltSmooth(Bitmap:TBitmap;TransTable:pointer);
procedure bmFlipX(Bitmap:TBitmap);
procedure bmFlipY(Bitmap:TBitmap);
procedure bmAnd(Bitmap1,Bitmap2:TBitmap);
procedure bmAndN(Bitmap1,Bitmap2:TBitmap);
procedure bmOr(Bitmap1,Bitmap2:TBitmap);
procedure bmXor(Bitmap1,Bitmap2:TBitmap);
procedure bmNot(Bitmap:TBitmap);

implementation
uses SinTablL;

procedure bmRescaleX(OldBitmap,NewBitmap:TBitmap);
var Factor,OldColumn,NewColumn:dword;
    Row:word;
begin
 Factor:=(OldBitmap.Width shl 16) div NewBitmap.Width;
 for Row:=1 to OldBitmap.Height do begin
  OldColumn:=0;
  for NewColumn:=0 to NewBitmap.Width-1 do begin
   Mem[NewBitmap._Body+NewColumn]:=Mem[OldBitmap._Body+(OldColumn shr 16)];
   Inc(OldColumn,Factor);
  end;
  Inc(OldBitmap.Body,OldBitmap.Width);
  Inc(NewBitmap.Body,NewBitmap.Width);
 end;
end;

procedure bmRescaleY(OldBitmap,NewBitmap:TBitmap);
var Factor,OldRow,NewRow:dword;
begin
 Factor:=(OldBitmap.Height shl 16) div NewBitmap.Height;
 OldRow:=0;
 for NewRow:=0 to NewBitmap.Height-1 do begin
  Move((OldBitmap.Body+(OldRow shr 16)*OldBitmap.Width)^,
       (NewBitmap.Body+NewRow*OldBitmap.Width)^,OldBitmap.Width);
  Inc(OldRow,Factor);
 end;
end;

procedure bmRescaleXY(OldBitmap,NewBitmap:TBitmap);
var FactorX,FactorY,
    OldColumn,NewColumn,
    OldRow,NewRow:dword;
    OldBmp:dword;
begin
 FactorY:=(OldBitmap.Height shl 16) div NewBitmap.Height;
 FactorX:=(OldBitmap.Width shl 16) div NewBitmap.Width;
 OldRow:=0; OldBmp:=OldBitmap._Body;
 for NewRow:=1 to NewBitmap.Height do begin
  OldColumn:=0;
  for NewColumn:=0 to NewBitmap.Width-1 do begin
   Mem[NewBitmap._Body+NewColumn]:=Mem[OldBmp+(OldColumn shr 16)];
   Inc(OldColumn,FactorX);
  end;
  Inc(OldRow,FactorY);
  OldBmp:=OldBitmap._Body+OldBitmap.Width*(OldRow shr 16);
  Inc(NewBitmap.Body,NewBitmap.Width);
 end;
end;

procedure bmShiftLine(Bitmap:TBitmap;Line,ShiftValue:word);
asm
 std
 movzx eax,Line
 movzx ecx,Bitmap.Width
 mul ecx                (* eax = line start offset *)
 mov esi,Bitmap.Body
 add esi,eax            (* esi = line start address *)
 add esi,ecx            (* esi = line end address + 1 *)
 dec esi                (* esi = line end address *)
 mov edi,esi
@@1:
 push esi
 lodsb                  (* take one byte from line end, dec esi *)
 movzx ecx,Bitmap.Width
 dec ecx                (* ecx = bitmap width - 1 *)
 rep movsb              (* move line, edi = line start offset *)
 stosb                  (* put saved byte to line start *)
 pop esi
 mov edi,esi
 dec ShiftValue         (* repeat until shifted *)
 jnz @@1
 cld
end;

procedure bmShiftLineZ(Bitmap:TBitmap;Line,ShiftValue:word;FillColor:byte);
asm
 std
 movzx eax,Line
 movzx ecx,Bitmap.Width
 mul ecx                (* eax = line start offset *)
 mov esi,Bitmap.Body
 add esi,eax            (* esi = line start address *)
 add esi,ecx            (* esi = line end address + 1 *)
 dec esi                (* esi = line end address *)
 mov edi,esi
 dec esi
@@1:
 push esi
 push edi
 movzx ecx,Bitmap.Width
 dec ecx                (* ecx = bitmap width - 1 *)
 rep movsb              (* move line, edi = line start offset *)
 mov al,FillColor
 mov [edi],al           (* put fill color to line start *)
 pop edi
 pop esi
 dec ShiftValue         (* repeat until shifted *)
 jnz @@1
 cld
end;

procedure bmShiftColumn(Bitmap:TBitmap;Column,ShiftValue:word);
asm
 movzx eax,Bitmap.Width
 movzx ecx,Bitmap.Height
 dec ecx
 mul ecx
 movzx ebx,Column
 add eax,ebx            (* eax = offset of last column's pixel *)

 movzx edx,Bitmap.Width
 mov esi,Bitmap.Body
 add esi,eax            (* esi = address of last column's pixel *)
 mov edi,esi

@@1:
 push esi
 push ecx               (* ecx = height - 1 *)
 mov bl,[esi]           (* take one byte from column end *)
 sub esi,edx

@@2:
 mov al,[esi]
 mov [edi],al
 sub esi,edx
 sub edi,edx
 dec ecx
 jnz @@2

 mov [edi],bl
 pop ecx
 pop esi
 mov edi,esi
 dec ShiftValue         (* repeat until shifted *)
 jnz @@1
end;

procedure bmShiftColumnZ(Bitmap:TBitmap;Column,ShiftValue:word;FillColor:byte);
asm
 movzx eax,Bitmap.Width
 movzx ecx,Bitmap.Height
 dec ecx
 mul ecx
 movzx ebx,Column
 add eax,ebx            (* eax = offset of last column's pixel *)

 movzx edx,Bitmap.Width
 mov esi,Bitmap.Body
 add esi,eax            (* esi = address of last column's pixel *)
 mov edi,esi
 sub esi,edx

@@1:
 push esi
 push edi
 push ecx               (* ecx = height - 1 *)

@@2:
 mov al,[esi]
 mov [edi],al
 sub esi,edx
 sub edi,edx
 dec ecx
 jnz @@2

 mov bl,FillColor
 mov [edi],bl
 pop ecx
 pop edi
 pop esi
 dec ShiftValue         (* repeat until shifted *)
 jnz @@1
end;

procedure bmRotate(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle:word);
var xscale,yscale,xc,yc:longint;
    scanline:word;
    x,y:Integer;
    tempx,tempy:integer;
    xlong,ylong:Longint;
begin
 xscale:=SineL[angle];
 yscale:=CosineL[angle];
 xc:=(OldBitmap.width shr 1 -
     (OldBitmap.width shr 1 - cx)) shl 16 -
     (cx*yscale+cy*xscale);
 yc:=(OldBitmap.height shr 1 -
     (OldBitmap.height shr 1 - cy)) shl 16 -
     (cy*yscale-cx*xscale);
 scanline:=0;
 for y:=0 to OldBitmap.height-1 do begin
  xlong:=xc;
  ylong:=yc;
  for x:=0 to OldBitmap.width-1 do begin
   tempx:=xlong shr 16;
   tempy:=ylong shr 16;
   if tempx<0 then tempx:=tempx mod OldBitmap.Width+OldBitmap.Width;
   if tempx>=OldBitmap.Width then tempx:=tempx mod OldBitmap.Width;
   if tempy<0 then tempy:=tempy mod OldBitmap.Height+OldBitmap.Height;
   if tempy>=OldBitmap.Height then tempy:=tempy mod OldBitmap.Height;
   Mem[NewBitmap._body+scanline+x]:=
    Mem[OldBitmap._body+tempy*OldBitmap.width+tempx];
   Inc(xlong,yscale);
   Dec(ylong,xscale);
  end;
  Inc(scanline,OldBitmap.Width);
  Inc(xc,xscale);
  Inc(yc,yscale);
 end;
end;

procedure bmRotateZ(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle:word;FillColor:byte);
var xscale,yscale,xc,yc:longint;
    scanline:word;
    x,y:Integer;
    tempx,tempy:integer;
    xlong,ylong:Longint;
begin
 xscale:=SineL[angle];
 yscale:=CosineL[angle];
 xc:=(OldBitmap.width shr 1 -
     (OldBitmap.width shr 1 - cx)) shl 16 -
     (cx*yscale+cy*xscale);
 yc:=(OldBitmap.height shr 1 -
     (OldBitmap.height shr 1 - cy)) shl 16 -
     (cy*yscale-cx*xscale);
 scanline:=0;
 for y:=0 to OldBitmap.height-1 do begin
  xlong:=xc;
  ylong:=yc;
  for x:=0 to OldBitmap.width-1 do begin
   tempx:=xlong shr 16;
   tempy:=ylong shr 16;
   if (tempx<0) or
      (tempx>=OldBitmap.Width) or
      (tempy<0) or
      (tempy>=OldBitmap.Height) then
       Mem[newbitmap._body+scanline+x]:=FillColor
   else
    Mem[newbitmap._body+scanline+x]:=
     Mem[OldBitmap._body+OldBitmap.width*tempy+tempx];
   Inc(xlong,yscale);
   Dec(ylong,xscale);
  end;
  Inc(scanline,OldBitmap.width);
  Inc(xc,xscale);
  Inc(yc,yscale);
 end;
end;

procedure bmRotScale(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle,ScaleFactor:word);
   (* ScaleFactor is OldSize/NewSize*256 *)
var xscale,yscale,xc,yc:longint;
    scanline:word;
    x,y:Integer;
    tempx,tempy:integer;
    xlong,ylong:Longint;
begin
 xscale:=SineL[angle]*ScaleFactor shr 8;
 yscale:=CosineL[angle]*ScaleFactor shr 8;
 xc:=(OldBitmap.width shr 1 -
     (OldBitmap.width shr 1 - cx)) shl 16 -
     (cx*yscale+cy*xscale);
 yc:=(OldBitmap.height shr 1 -
     (OldBitmap.height shr 1 - cy)) shl 16 -
     (cy*yscale-cx*xscale);
 scanline:=0;
 for y:=0 to OldBitmap.height-1 do begin
  xlong:=xc;
  ylong:=yc;
  for x:=0 to OldBitmap.width-1 do begin
   tempx:=xlong shr 16;
   tempy:=ylong shr 16;
   if tempx<0 then tempx:=tempx mod OldBitmap.Width+OldBitmap.Width;
   if tempx>=OldBitmap.Width then tempx:=tempx mod OldBitmap.Width;
   if tempy<0 then tempy:=tempy mod OldBitmap.Height+OldBitmap.Height;
   if tempy>=OldBitmap.Height then tempy:=tempy mod OldBitmap.Height;
   Mem[NewBitmap._body+scanline+x]:=
    Mem[OldBitmap._body+tempy*OldBitmap.width+tempx];
   Inc(xlong,yscale);
   Dec(ylong,xscale);
  end;
  Inc(scanline,OldBitmap.Width);
  Inc(xc,xscale);
  Inc(yc,yscale);
 end;
end;

procedure bmRotScaleZ(OldBitmap,NewBitmap:TBitmap;cx,cy,Angle,ScaleFactor:word;FillColor:byte);
var xscale,yscale,xc,yc:longint;
    scanline:word;
    x,y:Integer;
    tempx,tempy:integer;
    xlong,ylong:Longint;
begin
 xscale:=SineL[angle]*ScaleFactor shr 8;
 yscale:=CosineL[angle]*ScaleFactor shr 8;
 xc:=(OldBitmap.width shr 1 -
     (OldBitmap.width shr 1 - cx)) shl 16 -
     (cx*yscale+cy*xscale);
 yc:=(OldBitmap.height shr 1 -
     (OldBitmap.height shr 1 - cy)) shl 16 -
     (cy*yscale-cx*xscale);
 scanline:=0;
 for y:=0 to OldBitmap.height-1 do begin
  xlong:=xc;
  ylong:=yc;
  for x:=0 to OldBitmap.width-1 do begin
   tempx:=xlong shr 16;
   tempy:=ylong shr 16;
   if (tempx<0) or
      (tempx>=OldBitmap.Width) or
      (tempy<0) or
      (tempy>=OldBitmap.Height) then
       Mem[newbitmap._body+scanline+x]:=FillColor
   else
    Mem[newbitmap._body+scanline+x]:=
     Mem[OldBitmap._body+OldBitmap.width*tempy+tempx];
   Inc(xlong,yscale);
   Dec(ylong,xscale);
  end;
  Inc(scanline,OldBitmap.width);
  Inc(xc,xscale);
  Inc(yc,yscale);
 end;
end;

procedure bmHalf(OldBitmap,NewBitmap:TBitmap);
var i,j:word;
begin
 for i:=0 to OldBitmap.Height shr 1-1 do
  for j:=0 to OldBitmap.Width shr 1-1 do
   Mem[NewBitmap._Body+i*NewBitmap.Width+j]:=
    Mem[OldBitmap._Body+(i shl 1)*OldBitmap.Width+(j shl 1)];
end;

procedure bmDouble(OldBitmap,NewBitmap:TBitmap);
var i,j:word;
begin
 for i:=0 to OldBitmap.Height shl 1-1 do
  for j:=0 to OldBitmap.Width shl 1-1 do
   Mem[NewBitmap._Body+i*NewBitmap.Width+j]:=
    Mem[OldBitmap._Body+(i shr 1)*OldBitmap.Width+(j shr 1)];
end;

procedure bmHalfSmooth(OldBitmap,NewBitmap:TBitmap;TransTable:pointer);
var i,j:word;
begin
 for i:=0 to NewBitmap.Height-1 do
  for j:=0 to NewBitmap.Width-1 do
   Mem[NewBitmap._Body+i*NewBitmap.Width+j]:=
    Mem[dword(TransTable)+
    dword(Mem[OldBitmap._Body+(i shl 1+1)*OldBitmap.Width+(j shl 1)]) shl 8+
    Mem[dword(TransTable)+
    dword(Mem[OldBitmap._Body+(i shl 1)*OldBitmap.Width+(j shl 1)]) shl 8+
    Mem[OldBitmap._Body+(i shl 1)*OldBitmap.Width+(j shl 1)+1]]];
end;

procedure bmDoubleInterlace(OldBitmap,NewBitmap:TBitmap);
var i,j:word;
begin
 for i:=0 to OldBitmap.Height-1 do
  for j:=0 to OldBitmap.Width shl 1-1 do
   Mem[NewBitmap._Body+(i shl 1)*NewBitmap.Width+j]:=
    Mem[OldBitmap._Body+i*OldBitmap.Width+(j shr 1)];
end;

procedure bmSmooth(Bitmap:TBitmap;TransTable:pointer);
asm
 movzx ebx,Bitmap.Width
 movzx eax,Bitmap.Height
 dec eax
 mul ebx
 dec eax
 mov ecx,eax

 mov esi,Bitmap.Body
 mov edi,TransTable

@@1:
 mov ah,[esi]

 mov dh,ah
 mov dl,[esi+1]
 mov al,[edi+edx]
 mov [esi],al

 mov dh,dl
 mov dl,[esi+ebx+1]
 mov al,[edi+edx]
 mov [esi+1],al

 mov dh,dl
 mov dl,[esi+ebx]
 mov al,[edi+edx]
 mov [esi+ebx+1],al

 mov dh,dl
 mov dl,ah
 mov al,[edi+edx]
 mov [esi+ebx],al

 inc esi
 loop @@1
end;

procedure bmAltSmooth(Bitmap:TBitmap;TransTable:pointer);
        (* Alternative smoothing algorithm *)
asm
 movzx ebx,Bitmap.Width
 movzx eax,Bitmap.Height
 dec eax
 mul ebx
 dec eax
 mov ecx,eax

 mov esi,Bitmap.Body
 mov edi,TransTable

@@1:
 mov dh,[esi]
 mov dl,[esi+ebx+1]
 mov al,[edi+edx]
 mov [esi],al

 inc esi
 loop @@1
end;

procedure bmFlipX(Bitmap:TBitmap);
var i,j:word;
    c:byte;
begin
 for i:=0 to Bitmap.Height-1 do
  for j:=0 to Bitmap.Width shr 1-1 do begin
   c:=Mem[Bitmap._Body+i*Bitmap.Width+j];
   Mem[Bitmap._Body+i*Bitmap.Width+j]:=
    Mem[Bitmap._Body+i*Bitmap.Width+Bitmap.Width-1-j];
   Mem[Bitmap._Body+i*Bitmap.Width+Bitmap.Width-1-j]:=c;
  end;
end;

procedure bmFlipY(Bitmap:TBitmap);
var i,j:word;
    c:byte;
begin
 for i:=0 to Bitmap.Width-1 do
  for j:=0 to Bitmap.Height shr 1-1 do begin
   c:=Mem[Bitmap._Body+j*Bitmap.Width+i];
   Mem[Bitmap._Body+j*Bitmap.Width+i]:=
    Mem[Bitmap._Body+(Bitmap.Height-1-j)*Bitmap.Width+i];
   Mem[Bitmap._Body+(Bitmap.Height-1-j)*Bitmap.Width+i]:=c;
  end;
end;

procedure bmAnd(Bitmap1,Bitmap2:TBitmap);
asm
 mov ax,Bitmap1.Width
 mul Bitmap1.Height
 mov cx,dx
 shl ecx,16
 mov cx,ax
 mov ebx,ecx
 shr ecx,2
 jecxz @@2
 mov edi,Bitmap1.Body
 mov esi,Bitmap2.Body
@@1:
 mov eax,[edi]
 and eax,[esi]
 stosd
 add esi,4
 loop @@1
@@2:
 mov ecx,ebx
 and ecx,3
@@3:
 mov al,[edi]
 and al,[esi]
 stosb
 inc esi
 loop @@3
end;

procedure bmAndN(Bitmap1,Bitmap2:TBitmap);
asm
 mov ax,Bitmap1.Width
 mul Bitmap1.Height
 mov cx,dx
 shl ecx,16
 mov cx,ax
 mov ebx,ecx
 shr ecx,2
 jecxz @@2
 mov edi,Bitmap1.Body
 mov esi,Bitmap2.Body
@@1:
 mov eax,[edi]
 and eax,[esi]
 not eax
 stosd
 add esi,4
 loop @@1
@@2:
 mov ecx,ebx
 and ecx,3
@@3:
 mov al,[edi]
 and al,[esi]
 not al
 stosb
 inc esi
 loop @@3
end;

procedure bmOr(Bitmap1,Bitmap2:TBitmap);
asm
 mov ax,Bitmap1.Width
 mul Bitmap1.Height
 mov cx,dx
 shl ecx,16
 mov cx,ax
 mov ebx,ecx
 shr ecx,2
 jecxz @@2
 mov edi,Bitmap1.Body
 mov esi,Bitmap2.Body
@@1:
 mov eax,[edi]
 or eax,[esi]
 stosd
 add esi,4
 loop @@1
@@2:
 mov ecx,ebx
 and ecx,3
@@3:
 mov al,[edi]
 or al,[esi]
 stosb
 inc esi
 loop @@3
end;

procedure bmXor(Bitmap1,Bitmap2:TBitmap);
asm
 mov ax,Bitmap1.Width
 mul Bitmap1.Height
 mov cx,dx
 shl ecx,16
 mov cx,ax
 mov ebx,ecx
 shr ecx,2
 jecxz @@2
 mov edi,Bitmap1.Body
 mov esi,Bitmap2.Body
@@1:
 mov eax,[edi]
 xor eax,[esi]
 stosd
 add esi,4
 loop @@1
@@2:
 mov ecx,ebx
 and ecx,3
@@3:
 mov al,[edi]
 xor al,[esi]
 stosb
 inc esi
 loop @@3
end;

procedure bmNot(Bitmap:TBitmap);
asm
 mov ax,Bitmap.Width
 mul Bitmap.Height
 mov cx,dx
 shl ecx,16
 mov cx,ax
 mov ebx,ecx
 shr ecx,2
 jecxz @@2
 mov esi,Bitmap.Body
@@1:
 not dword ptr [esi]
 add esi,4
 loop @@1
@@2:
 mov ecx,ebx
 and ecx,3
@@3:
 not byte ptr [esi]
 inc esi
 loop @@3
end;

end.
