(*******************************************)
(* VRg SuperVGA library                    *)
(* Font Library management utilities       *)
(*  v. 1.02 09.Jul.1999                    *)
(* Copyleft (C) 1998-1999 Virtual Research *)
(*******************************************)

unit fl;
{$S-,Q-,R-,I-,V-,B-,X+}

interface

type TFLHeader=record
 Signature:array[0..21] of char;
 NumFonts:dword;
end;

     TFontName=array[0..7] of char;
     TComment=array[0..69] of char;

     TDirEntry=record
 FontName:TFontName;
 Comment:TComment;
end;

     TFLDir=record
 NumEntries:dword;
 SizeInBytes:dword;
 Entries:array[dword] of TDirEntry;
end;
     PFLDir=^TFLDir;

procedure flCreate(FileName:string;var FL:file);
function  flGetDir(var FL:file):PFLDir;
procedure flGetFont(var FL:file;Name:TFontName;Buffer:pointer);
procedure flAddFont(var FL:file;Name:TFontName;Comment:TComment;Buffer:pointer);
procedure flDelFont(var FL:file;Name:TFontName);
procedure flDisposeDir(FLDir:PFLDir);

var flError:integer;

const flErrorMessage:array[1..4] of string[27]=(
      'Could not write to library',
      'Could not read from library',
      'Invalid library file format',
      'Font not found in library');


implementation
uses DOS;

procedure flCreate(FileName:string;var FL:file);
var Header:TFLHeader;
begin
 flError:=0;
 Assign(FL,FileName);
 Rewrite(FL,1);
 if IOResult<>0 then begin
  flError:=1;	(* disk write error *)
  Exit;
 end;
 Header.Signature:='VR group Font Library'#26;
 Header.NumFonts:=0;
 BlockWrite(FL,Header,SizeOf(TFLHeader));
 if IOResult<>0 then begin
  flError:=1;	(* disk write error *)
  Exit;
 end;
end;

function  GetHeader(var FL:file):TFLHeader;
begin
 flError:=0;
 Seek(FL,0);
 BlockRead(FL,Result,SizeOf(TFLHeader));
 if IOResult<>0 then begin
  flError:=2;	(* disk read error *)
  Exit;
 end;
 if Result.Signature<>'VR group Font Library'#26 then
  flError:=3;  (* not an FL *)
end;

function  flGetDir(var FL:file):PFLDir;
var Header:TFLHeader;
    DirSize:dword;
begin
 flError:=0;
 Header:=GetHeader(FL);
 if flError<>0 then Exit;
 DirSize:=Header.NumFonts*SizeOf(TDirEntry);
 GetMem(Result,DirSize+8);
 Result^.NumEntries:=Header.NumFonts;
 Result^.SizeInBytes:=DirSize+8;
 Seek(FL,Header.NumFonts*4096+SizeOf(TFLHeader));
 BlockRead(FL,Result^.Entries,DirSize);
 if IOResult<>0 then begin
  flError:=2;	(* disk read error *)
  Exit;
 end;
end;

function  GetIndex(FLDir:PFLDir;Name:TFontName):dword;
var i:dword;
    Found:boolean;
begin
 flError:=0;
 Found:=false;
 for i:=0 to FLDir^.NumEntries-1 do
  if FLDir^.Entries[i].FontName=Name then begin
   Result:=i; Found:=true; Break;
  end;
 if not Found then begin Result:=0; flError:=4; end;
end;

procedure flGetFont(var FL:file;Name:TFontName;Buffer:pointer);
var Index:dword; FLDir:PFLDir;
begin
 flError:=0;
 FLDir:=flGetDir(FL);
 if flError<>0 then begin
  flDisposeDir(FLDir);
  Exit;
 end;
 Index:=GetIndex(FLDir,Name);
 if flError<>0 then begin
  flDisposeDir(FLDir);
  Exit;
 end;
 Seek(FL,4096*Index+SizeOf(TFLHeader));
 BlockRead(FL,Buffer^,4096);
 if IOResult<>0 then begin
  flError:=2;	(* disk read error *)
  Exit;
 end;
end;

procedure flAddFont(var FL:file;Name:TFontName;Comment:TComment;Buffer:pointer);
var FLDir:PFLDir; NewEntry:TDirEntry; Header:TFLHeader;
begin
 flError:=0;
 Header:=GetHeader(FL);
 if flError<>0 then Exit;
 FLDir:=flGetDir(FL);
 if flError<>0 then begin
  flDisposeDir(FLDir);
  Exit;
 end;
 NewEntry.FontName:=Name;
 NewEntry.Comment:=Comment;
 Seek(FL,Header.NumFonts*4096+SizeOf(TFLHeader));
 BlockWrite(FL,Buffer^,4096);
 BlockWrite(FL,FLDir^.Entries,Header.NumFonts*SizeOf(TDirEntry));
 BlockWrite(FL,NewEntry,SizeOf(TDirEntry));
 Inc(Header.NumFonts);
 Seek(FL,0);
 BlockWrite(FL,Header,SizeOf(TFLHeader));
 flDisposeDir(FLDir);
 if IOResult<>0 then begin
  flError:=1;
  Exit;
 end;
end;

procedure DelEntry(FLDir:PFLDir;Index:dword);
begin
 Dec(FLDir^.NumEntries);
 Move(FLDir^.Entries[Index+1],FLDir^.Entries[Index],
      (FLDir^.NumEntries-Index)*SizeOf(TDirEntry));
end;

procedure flDisposeDir(FLDir:PFLDir);
begin
 FreeMem(FLDir,FLDir^.SizeInBytes);
end;

procedure flDelFont(var FL:file;Name:TFontName);
var Index:dword; FLDir:PFLDir; Header:TFLHeader;
    i,FontsToShift:dword;
    Buffer:array[1..4096] of byte;
begin
 flError:=0;
 Header:=GetHeader(FL);
 if Header.NumFonts=0 then Exit;
 if flError<>0 then Exit;
 FLDir:=flGetDir(FL);
 if flError<>0 then begin
  flDisposeDir(FLDir);
  Exit;
 end;
 Index:=GetIndex(FLDir,Name);
 if flError<>0 then begin
  flDisposeDir(FLDir);
  Exit;
 end;
 if Header.NumFonts>1 then begin
  FontsToShift:=Header.NumFonts-1-Index;
  for i:=1 to FontsToShift do begin
   Seek(FL,(Index+i)*4096+SizeOf(TFLHeader));
   BlockRead(FL,Buffer,4096);
   if IOResult<>0 then begin
    flError:=2;
    Exit;
   end;
   Seek(FL,(Index+i-1)*4096+SizeOf(TFLHeader));
   BlockWrite(FL,Buffer,4096);
   if IOResult<>0 then begin
    flError:=1;
    Exit;
   end;
  end;
 end
 else Seek(FL,SizeOf(TFLHeader));
 Dec(Header.NumFonts);
 DelEntry(FLDir,Index);
 BlockWrite(FL,FLDir^.Entries,Header.NumFonts*SizeOf(TDirEntry));
 Truncate(FL); Seek(FL,0);
 BlockWrite(FL,Header,SizeOf(TFLHeader));
 if IOResult<>0 then flError:=1;
end;

end.

