(*****************************************************)
(* Autodesk FLI animation player for TMT Pascal 2.0+ *)
(* Copyleft (C) 1998-1999 Virtual Research i. g.     *)
(* All Lefts Reserved                                *)
(*                                                   *)
(* Unit version : 1.21                               *)
(* Started      : 30.May.1998                        *)
(* Finished     : 01.Jun.1998                        *)
(* Last updated : 16.Jul.1999                        *)
(* System       : 32-bit DPMI                        *)
(*****************************************************)

unit FLI32;
{$S-,R-,Q-,V-,B-,X+}

interface

       (* Flags *)
 const Version='1.21';
       fliInfiniteLoop:boolean=false;   (* Loop flick for ever *)
       fliCheckKeyPress:boolean=true;   (* Exit on <Esc> *)
       fliLoopTimes:word=1;             (* Loop some times *)
       
       fliUserPreFrame:procedure=nil;   (* Will be called before each frame *)
       fliUserPostFrame:procedure=nil;  (* Will be called after each frame *)

 var fliVFB:pointer;                    (* Virtual framebuffer address *)

 procedure fliPlayFromMemory(var Flick);  (* Play pre-loaded flick *)
 procedure fliPlayFromFile(FileName:string;FileOffset:dword);

implementation
uses LFB256;

var Timer:dword absolute $46C;
    Ticks:dword;

procedure TimerReset;
begin
 Ticks:=Timer;
end;

function TimerRead:dword;
begin
 TimerRead:=Timer-Ticks;
end;

{$I flick.inc}

var   FLI_Header:TFlickHeader;
      VFB:array[0..63999] of byte;      (* Virtual framebuffer *)
      VFBPos:word;                      (* Relative position in VFB *)
      Palette:TPalette;
      Position:dword;
      FlickPtr:pointer;

const _FLI_COLOR=11;
      _FLI_LC=12;
      _FLI_BLACK=13;
      _FLI_BRUN=15;
      _FLI_COPY=16;

procedure GetData(var Destination;Count:dword);
begin
 Move((FlickPtr+Position)^,Destination,Count);
 Inc(Position,Count);
end;

procedure FLI_COLOR;
var Packets,i:word; Skip:byte;
const PalettePosition:byte=0; Count:word=0; Dummy:word=0;
begin
 GetData(Packets,2);
 for i:=1 to Packets do begin
  GetData(Skip,1);
  GetData(Count,1);
  Inc(Dummy,Count);
  if Count=0 then Count:=256;
  Inc(PalettePosition,Skip);
  GetData(Palette[PalettePosition],Count*3);
  Inc(PalettePosition,Count);
 end;
 if Odd(Dummy*3) then GetData(Skip,1);
 _SetPalette(Palette);
end;

procedure FLI_LC;
var SkipLines,ChangeLines,i:word; Packets,j:byte;
    SkipPix,PixData:byte; ChangePix:shortint;
begin
 VFBPos:=0;
 GetData(SkipLines,2);
 GetData(ChangeLines,2);
 for i:=1 to ChangeLines do begin
  VFBPos:=(SkipLines+i-1)*320;
  GetData(Packets,1);
  for j:=1 to Packets do begin
   GetData(SkipPix,1);
   Inc(VFBPos,SkipPix);
   GetData(ChangePix,1);
   if ChangePix>=0 then begin
    GetData(VFB[VFBPos],ChangePix);
    Inc(VFBPos,ChangePix);
   end
   else begin
    GetData(PixData,1);
    FillChar(VFB[VFBPos],-ChangePix,PixData);
    Inc(VFBPos,-ChangePix);
   end;
  end;
 end;
end;

procedure FLI_BLACK;
begin
 FillChar(VFB,64000,0);
end;

procedure FLI_BRUN;
var i:word; Packets,j:byte;
    PixData:byte; ChangePix:shortint;
begin
 VFBPos:=0;
 for i:=1 to 200 do begin
  GetData(Packets,1);
  for j:=1 to Packets do begin
   GetData(ChangePix,1);
   if ChangePix<0 then begin
    GetData(VFB[VFBPos],-ChangePix);
    Inc(VFBPos,-ChangePix);
   end
   else begin
    GetData(PixData,1);
    FillChar(VFB[VFBPos],ChangePix,PixData);
    Inc(VFBPos,ChangePix);
   end;
  end;
 end;
end;

procedure FLI_COPY;
begin
 GetData(VFB,64000);
end;


procedure fliPlayFromMemory(var Flick);
var Loop,FrameCount,ChunkCount:word;
    FrameHeader:TFrameHeader;
    ChunkSize:dword; ChunkType:word;
    FrameStart:dword; Delay:dword;

begin
 Position:=0;FlickPtr:=@Flick;
 GetData(FLI_Header,128);
 if FLI_Header.Signature<>$AF11 then Exit; (* Not a FLI *)
 if (FLI_Header.Width<>320) or (FLI_Header.Height<>200) then Exit;
                                                          (* Wrong size *)
 Delay:=FLI_Header.Speed shr 2;
 repeat
  for Loop:=1 to fliLoopTimes do begin
   FrameStart:=128;
   for FrameCount:=1 to FLI_Header.Frames do begin
    TimerReset;
    Position:=FrameStart;
    GetData(FrameHeader,16);
    Inc(FrameStart,FrameHeader.Size);
    if FrameHeader.ID=$F1FA then begin
     for ChunkCount:=1 to FrameHeader.Chunks do begin
      GetData(ChunkSize,4);
      GetData(ChunkType,2);
      Case ChunkType of
       _FLI_COLOR:FLI_COLOR;
       _FLI_LC:FLI_LC;
       _FLI_BLACK:FLI_BLACK;
       _FLI_BRUN:FLI_BRUN;
       _FLI_COPY:FLI_COPY;
      end;
      if @fliUserPreFrame<>nil then fliUserPreFrame;
      if ChunkType<>_FLI_COLOR then Move(VFB,pointer($A0000)^,64000);
     end;  
    end;
    if fliCheckKeyPress then if Port[$60]=1 then begin
     MemW[$41C-2]:=MemW[$41C];  (* Flush keyboard *)
     Exit; 
    end;
    if @fliUserPostFrame<>nil then fliUserPostFrame;
    while TimerRead<Delay do;
   end;
  end; 
 until (not fliInfiniteLoop);
end;

procedure fliPlayFromFile(FileName:string;FileOffset:dword);
var Loop,FrameCount,ChunkCount:word;
    FrameHeader:TFrameHeader;
    ChunkSize:dword; ChunkType:word;
    FlickFile:file; Delay:dword;
    PFB:pointer;        (* Packed frame buffer *)

begin
 GetMem(PFB,$FFF0);
 FlickPtr:=PFB;
 Assign(FlickFile,FileName);
 Reset(FlickFile,1); Seek(FlickFile,FileOffset);
 BlockRead(FlickFile,FLI_Header,128);
 if FLI_Header.Signature<>$AF11 then Exit; (* Not a FLI *)
 if (FLI_Header.Width<>320) or (FLI_Header.Height<>200) then Exit;
                                                         (* Wrong size *)
 Delay:=FLI_Header.Speed shr 2;
 repeat
  for Loop:=1 to fliLoopTimes do begin
   Seek(FlickFile,FileOffset+128);
   for FrameCount:=1 to FLI_Header.Frames do begin
    TimerReset;
    BlockRead(FlickFile,FrameHeader,16);
    BlockRead(FlickFile,PFB^,FrameHeader.Size-16);
    Position:=0;
    if FrameHeader.ID=$F1FA then begin
     for ChunkCount:=1 to FrameHeader.Chunks do begin
      GetData(ChunkSize,4);
      GetData(ChunkType,2);
      Case ChunkType of
       _FLI_COLOR:FLI_COLOR;
       _FLI_LC:FLI_LC;
       _FLI_BLACK:FLI_BLACK;
       _FLI_BRUN:FLI_BRUN;
       _FLI_COPY:FLI_COPY;
      end;
      if @fliUserPreFrame<>nil then fliUserPreFrame;
      if ChunkType<>_FLI_COLOR then Move(VFB,pointer($A0000)^,64000);
     end;  
    end;
    if fliCheckKeyPress then if Port[$60]=1 then begin
     MemW[$41C-2]:=MemW[$41C];
     Exit; 
    end;
    if @fliUserPostFrame<>nil then fliUserPostFrame;
    while TimerRead<Delay do;
   end;
  end; 
 until (not fliInfiniteLoop);
 FreeMem(PFB,$FFF0);
 Close(FlickFile);
end;

begin
 fliVFB:=@VFB;
end.
