(**********************************************************)
(* Autodesk FLI/FLC animation decoder for TMT Pascal 2.0+ *)
(* Copyleft (C) 1999 Virtual Research                     *)
(*                                                        *)
(* Unit version : 1.00                                    *)
(* Last updated : 15.Jul.1999                             *)
(* System       : 32-bit DPMI                             *)
(**********************************************************)

unit Flick;
{$S-,R-,Q-,V-,B-,X+}

interface
uses LFB256;

 const Version='1.00';

{$I flick.inc}

 type TFlick=object
  constructor Init;
  function GetInfo:TFlickHeader;virtual;
  function GetVFB:pointer;virtual;
  function GetPalette:TPalette;virtual;
  function PaletteChanged:boolean;virtual;
  procedure DecodeFrame;virtual;
  function GetCurrentFrameNumber:word;virtual;
  destructor Done;virtual;
 private
  FlickHeader:TFlickHeader;
  FrameHeader:TFrameHeader;
  Palette:TPalette;
  PaletteChangeFlag:boolean;
  CurrentFrame:pointer;
  FramePosition:dword;
  FrameCount:word;
  VFB:pointer;
  procedure GetData(var Destination;Count:dword);virtual;
  procedure FLI_COLOR;virtual;
  procedure FLI_LC;virtual;
  procedure FLI_BLACK;virtual;
  procedure FLI_BRUN;virtual;
  procedure FLI_COPY;virtual;
  procedure FLI_DELTA;virtual;
  procedure FLI_256_COLOR;virtual;
 end;
      PFlick=^TFlick;

      TFlickFile=object(TFlick)
  constructor Init(FileName:string;FileOffset:dword);
  procedure DecodeFrame;virtual;
  destructor Done;virtual;
 private
  FlickFile:file;
  PFB:pointer;
 end;
      PFlickFile=^TFlickFile;

      TFlickBuffer=object(TFlick)
  constructor Init(Buffer:pointer);
  procedure DecodeFrame;virtual;
  destructor Done;virtual;
 private
  FlickBuffer:pointer;
 end;
      PFlickBuffer=^TFlickBuffer;


implementation

const _FLI_COLOR=11;
      _FLI_LC=12;
      _FLI_BLACK=13;
      _FLI_BRUN=15;
      _FLI_COPY=16;
      _FLI_DELTA=7;
      _FLI_256_COLOR=4;

procedure TFlick.GetData(var Destination;Count:dword);
begin
 Move((CurrentFrame+FramePosition)^,Destination,Count);
 Inc(FramePosition,Count);
end;

function TFlick.PaletteChanged:boolean;
begin
 Result:=PaletteChangeFlag;
end;

function TFlick.GetCurrentFrameNumber:word;
begin
 Result:=FrameCount;
end;

procedure TFlick.FLI_COLOR;
var Packets,i:word; Skip:byte;
const PalettePosition:byte=0;
      Count:word=0;
      Dummy:word=0;
begin
 GetData(Packets,2);
 for i:=1 to Packets do begin
  GetData(Skip,1);
  GetData(Count,1);
  Inc(Dummy,Count);
  if Count=0 then Count:=256;
  Inc(PalettePosition,Skip);
  GetData(Palette[PalettePosition],Count*3);
  Inc(PalettePosition,Count);
 end;
 if Odd(Dummy*3) then GetData(Skip,1);
end;

procedure TFlick.FLI_LC;
var SkipLines,ChangeLines,i:word;
    Packets,j:byte;
    SkipPix,PixData:byte;
    ChangePix:shortint;
    VFBPosition:dword;
begin
 VFBPosition:=0;
 GetData(SkipLines,2);
 GetData(ChangeLines,2);
 for i:=1 to ChangeLines do begin
  VFBPosition:=(SkipLines+i-1)*dword(FlickHeader.Width);
  GetData(Packets,1);
  for j:=1 to Packets do begin
   GetData(SkipPix,1);
   Inc(VFBPosition,SkipPix);
   GetData(ChangePix,1);
   if ChangePix>=0 then begin
    GetData((VFB+VFBPosition)^,ChangePix);
    Inc(VFBPosition,ChangePix);
   end
   else begin
    GetData(PixData,1);
    FillChar((VFB+VFBPosition)^,-ChangePix,PixData);
    Inc(VFBPosition,-ChangePix);
   end;
  end;
 end;
end;

procedure TFlick.FLI_BLACK;
begin
 FillChar(VFB^,dword(FlickHeader.Width)*FlickHeader.Height,0);
end;

procedure TFlick.FLI_BRUN;
var i:word; 
    Packets,j:byte;
    PixData:byte; 
    ChangePix:shortint;
    VFBPosition:dword;
begin
 VFBPosition:=0;
 for i:=1 to FlickHeader.Height do begin
  GetData(Packets,1);
  for j:=1 to Packets do begin
   GetData(ChangePix,1);
   if ChangePix<0 then begin
    GetData((VFB+VFBPosition)^,-ChangePix);
    Inc(VFBPosition,-ChangePix);
   end
   else begin
    GetData(PixData,1);
    FillChar((VFB+VFBPosition)^,ChangePix,PixData);
    Inc(VFBPosition,ChangePix);
   end;
  end;
 end;
end;

procedure TFlick.FLI_COPY;
begin
 GetData(VFB,dword(FlickHeader.Width)*FlickHeader.Height);
end;


procedure TFlick.FLI_DELTA;
var NumLines,Packets,j:integer;
    SkipPix:byte;
    PixData:word;
    ChangePix:shortint;
    VFBPosition,LineStart:dword;
    VFBPtr:pointer;
begin
 GetData(NumLines,2);
 LineStart:=0;
 while NumLines>0 do begin
  VFBPosition:=LineStart;
  GetData(Packets,2);
  if Packets<0 then Inc(LineStart,-Packets*dword(FlickHeader.Width))
  else begin
   for j:=1 to Packets do begin
    GetData(SkipPix,1);
    Inc(VFBPosition,SkipPix);
    GetData(ChangePix,1);
    if ChangePix>=0 then begin
     GetData((VFB+VFBPosition)^,ChangePix*2);
     Inc(VFBPosition,ChangePix*2);
    end
    else begin
     ChangePix:=-ChangePix;
     GetData(PixData,2);
     VFBPtr:=VFB+VFBPosition;
     asm
      mov ax,PixData
      movzx ecx,ChangePix
      mov edi,VFBPtr
      rep stosw
     end;
     Inc(VFBPosition,ChangePix*2);
    end;
   end;
   Inc(LineStart,FlickHeader.Width);
   Dec(NumLines);
  end;
 end;
end;

procedure TFlick.FLI_256_COLOR;
begin
 FLI_COLOR;
 _Convert8bppcTo6bppc(Palette);
end;

constructor TFlick.Init;
begin
 Fail;  (* This is pure prototype class, so it cannot be initialized *)
end;


function TFlick.GetInfo:TFlickHeader;
begin
 Result:=FlickHeader;
end;

function TFlick.GetVFB:pointer;
begin
 Result:=VFB;
end;

function TFlick.GetPalette:TPalette;
begin
 Result:=Palette;
end;

procedure TFlick.DecodeFrame;
var ChunkCount:word;
    ChunkSize:dword; ChunkType:word;
begin
 PaletteChangeFlag:=false;
 if FrameHeader.ID=$F1FA then begin
  for ChunkCount:=1 to FrameHeader.Chunks do begin
   GetData(ChunkSize,4);
   GetData(ChunkType,2);
   if (ChunkType=_FLI_COLOR) or (ChunkType=_FLI_256_COLOR) then
    PaletteChangeFlag:=true;
   Case ChunkType of
    _FLI_COLOR    : FLI_COLOR;
    _FLI_LC       : FLI_LC;
    _FLI_BLACK    : FLI_BLACK;
    _FLI_BRUN     : FLI_BRUN;
    _FLI_COPY     : FLI_COPY;
    _FLI_DELTA    : FLI_DELTA;
    _FLI_256_COLOR: FLI_256_COLOR;
    else Inc(FramePosition,ChunkSize-6)
   end;
  end;
 end;
end;

destructor TFlick.Done;
begin
end;

constructor TFlickFile.Init(FileName:string;FileOffset:dword);
var OldMode:word;
    Dummy:dword;
begin
 Assign(FlickFile,FileName);
 OldMode:=FileMode;
 Filemode:=0;
{$I-}
 Reset(FlickFile,1);
 Seek(FlickFile,FileOffset);
{$I+}
 FileMode:=OldMode;
 if IOResult<>0 then Fail;
 BlockRead(FlickFile,FlickHeader,SizeOf(TFlickHeader),Dummy);
 if MaxAvail<dword(FlickHeader.Width)*FlickHeader.Height*2 then begin
  Close(FlickFile);
  Fail;
 end;
 if FlickHeader.Signature=$AF11 then FlickHeader.Speed:=FlickHeader.Speed*14;
 GetMem(VFB,dword(FlickHeader.Width)*FlickHeader.Height);
 FillChar(VFB^,dword(FlickHeader.Width)*FlickHeader.Height,0);
 GetMem(PFB,dword(FlickHeader.Width)*FlickHeader.Height+4096);
 BlockRead(FlickFile,FrameHeader,SizeOf(TFrameHeader),Dummy);
 BlockRead(FlickFile,PFB^,FrameHeader.Size-SizeOf(TFrameHeader),Dummy);
 CurrentFrame:=PFB;
 FramePosition:=0;
 FrameCount:=0;
end;

procedure TFlickFile.DecodeFrame;
var Dummy:dword;
    ID:word;
begin
 repeat
  ID:=FrameHeader.ID; 
  inherited DecodeFrame;
  if ID=$F1FA then Inc(FrameCount);
  if FrameCount=FlickHeader.Frames then begin
   Seek(FlickFile,SizeOf(TFlickHeader));
   FrameCount:=0;
  end;
  BlockRead(FlickFile,FrameHeader,SizeOf(TFrameHeader),Dummy);
  BlockRead(FlickFile,PFB^,FrameHeader.Size-SizeOf(TFrameHeader),Dummy);
  CurrentFrame:=PFB;
  FramePosition:=0;
 until ID=$F1FA;    (* Skip irrelevant frames *)
end;

destructor TFlickFile.Done;
begin
 Close(FlickFile);
 FreeMem(VFB,dword(FlickHeader.Width)*FlickHeader.Height);
 FreeMem(PFB,dword(FlickHeader.Width)*FlickHeader.Height+4096);
end;


constructor TFlickBuffer.Init(Buffer:pointer);
begin
 FlickBuffer:=Buffer;
 Move(Buffer^,FlickHeader,SizeOf(TFlickHeader));
 if FlickHeader.Signature=$AF11 then FlickHeader.Speed:=FlickHeader.Speed*14;
 if MaxAvail<dword(FlickHeader.Width)*FlickHeader.Height then Fail;
 GetMem(VFB,dword(FlickHeader.Width)*FlickHeader.Height);
 FillChar(VFB^,dword(FlickHeader.Width)*FlickHeader.Height,0);
 CurrentFrame:=FlickBuffer+SizeOf(TFlickHeader);
 Move(CurrentFrame^,FrameHeader,SizeOf(TFrameHeader));
 Inc(CurrentFrame,SizeOf(TFrameHeader));
 FramePosition:=0;
 FrameCount:=0;
end;

procedure TFlickBuffer.DecodeFrame;
var ID:word;
begin
 repeat
  ID:=FrameHeader.ID;
  inherited DecodeFrame;
  if ID=$F1FA then Inc(FrameCount);
  if FrameCount=FlickHeader.Frames then begin
   CurrentFrame:=FlickBuffer+SizeOf(TFlickHeader);
   FrameCount:=0;
  end
  else Inc(CurrentFrame,FrameHeader.Size-SizeOf(TFrameHeader));
  Move(CurrentFrame^,FrameHeader,SizeOf(TFrameHeader));
  Inc(CurrentFrame,SizeOf(TFrameHeader));
  FramePosition:=0;
 until ID=$F1FA; (* Skip irrelevant frames *)
end;

destructor TFlickBuffer.Done;
begin
 FreeMem(VFB,dword(FlickHeader.Width)*FlickHeader.Height);
end;

end.
