(************************************************************)
(* Drawing routines with translucency fx for LFB256 library *)
(* Copyleft (C) 1998-1999 Virtual Research                  *)
(************************************************************)
(* Unit version : 1.2                                       *)
(* Last updated : 03.Jul.1999                               *)
(************************************************************)

unit trans256;
{$S-,Q-,R-,V-,B-,X+}

interface

const Version='1.2';

type TTransTable=array[0..255,0..255] of byte;
                  (* First index is background, second is foreground color *)
     PTransTable=^TTransTable;


 procedure transSetTable(TransTable:PTransTable);

   (* Drawing routines *)
   (********************)

 procedure transPutPixel(x,y:word;Color:byte);
 procedure transPutTransparentPixel(x,y:word;Color:byte);
 procedure transLine(x1,y1,x2,y2:word;Color:byte);
 procedure transMaskedLine(x1,y1,x2,y2:word;Color:byte;Mask:word);
 procedure transHLine(x1,x2,y:word;Color:byte);
 procedure transVLine(x,y1,y2:word;Color:byte);
 procedure transRectangle(x1,y1,x2,y2:word;Color:byte);
 procedure transBar(x1,y1,x2,y2:word;Color:byte);
 procedure transEllipse(xc,yc,rx,ry:word;Color:byte);
 procedure transFilledEllipse(xc,yc,rx,ry:word;Color:byte);
 procedure transTriangle(x1,y1,x2,y2,x3,y3:word;Color:byte);
 procedure transFillTriangle(x1,y1,x2,y2,x3,y3:integer;Color:byte);
 procedure transFillPoly(var VertexArray;NumOfVertexes,Color:byte);

   (* The same with clipping *)
   (**************************)

 procedure _transPutPixel(x,y:integer;Color:byte);
 procedure _transPutTransparentPixel(x,y:integer;Color:byte);
 procedure _transLine(x1,y1,x2,y2:integer;Color:byte);
 procedure _transMaskedLine(x1,y1,x2,y2:integer;Color:byte;Mask:word);
 procedure _transHLine(x1,x2,y:integer;Color:byte);
 procedure _transVLine(x,y1,y2:integer;Color:byte);
 procedure _transRectangle(x1,y1,x2,y2:integer;Color:byte);
 procedure _transBar(x1,y1,x2,y2:integer;Color:byte);
 procedure _transEllipse(xc,yc:integer;rx,ry:word;Color:byte);
 procedure _transFilledEllipse(xc,yc:integer;rx,ry:word;Color:byte);
 procedure _transTriangle(x1,y1,x2,y2,x3,y3:integer;Color:byte);
 procedure _transFillTriangle(x1,y1,x2,y2,x3,y3:integer;Color:byte);
 procedure _transFillPoly(var VertexArray;NumOfVertexes,Color:byte);

   (* Getting/putting bitmaps *)
   (***************************)

 procedure transPutImage(x,y,Width,Height:word;Address:pointer);
 procedure transPutTransparentImage(x,y,Width,Height:word;Address:pointer);

   (* The same with clipping *)
   (**************************)

 procedure _transPutImage(x,y:integer;Width,Height:word;Address:pointer);
 procedure _transPutTransparentImage(x,y:integer;Width,Height:word;Address:pointer);

   (* Text writing *)
   (****************)

 procedure transWriteText(x,y:word;S:string;Color:byte);
 procedure _transWriteText(x,y:word;S:string;Color:byte);

implementation
uses LFB256;

const _Table:PTransTable=nil;

procedure transSetTable(TransTable:PTransTable);
begin
 _Table:=TransTable;
end;

procedure transPutPixel(x,y:word;Color:byte);assembler;
asm
 mov edi,_FrameBuffer           
 movzx eax,y
 movzx edx,_LogicalScreenWidth  
 mul edx                        
 mov dx,x
 add eax,edx                    
 add edi,eax                    
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al                   
end;

procedure transPutTransparentPixel(x,y:word;Color:byte);assembler;
asm
 mov dl,Color
 cmp dl,_TransparentColor
 je @@Exit
 mov edi,_FrameBuffer
 movzx eax,y
 movzx edx,_LogicalScreenWidth
 mul edx
 mov dx,x
 add eax,edx
 add edi,eax
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al
@@Exit:
end;

procedure transLine(x1,y1,x2,y2:word;Color:byte);assembler;
var d,_dx,_dy,aincr,bincr,yincr:integer;
asm
 mov ax,x1
 cmp ax,x2
 je @@Vertical
 jbe @@1                
 mov bx,x2
 mov x1,bx
 mov x2,ax              
 mov ax,y1
 mov bx,y2
 mov y1,bx
 mov y2,ax              
@@1:
 mov cx,_LogicalScreenWidth
 mov ax,y1
 mov bx,y2
 cmp bx,ax
 ja @@2                 
 neg cx                 
@@2:
 mov yincr,cx            
 sub bx,ax              
 jns @@3
 neg bx
@@3:
 mov _dy,bx             
 mov ax,x2
 sub ax,x1
 mov _dx,ax             
 cmp ax,bx
 jb @@45                
 shl bx,1
 mov bincr,bx           
 sub bx,ax
 mov d,bx               
 mov ax,_dy
 sub ax,_dx
 jmp @@Continue
@@45:                   
 shl ax,1
 mov bincr,ax           
 sub ax,bx
 mov d,ax               
 mov ax,_dx
 sub ax,_dy
@@Continue:
 sal ax,1
 mov aincr,ax           

 movzx edx,_LogicalScreenWidth
 movzx eax,y1
 mul edx
 mov edi,_FrameBuffer
 add edi,eax
 mov dx,x1
 add edi,edx
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al          

 movsx edx,yincr
 mov ax,_dx
 cmp ax,_dy
 jb @@LessThen45

@@6:
 inc edi                
 inc x1
 mov ax,d
 cmp ax,0
 js @@4                 
 add edi,edx            
 add ax,aincr           
 jmp @@5
@@4:                    
 add ax,bincr           
@@5:
 mov d,ax
 movzx ebx,Color
 mov bh,[ebx]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al           

 mov ax,x2
 cmp ax,x1
 ja @@6                 
 jmp @@Exit

@@LessThen45:
 xor edx,0
 jns @@Positive
 mov cx,-1
 jmp @@6_
@@Positive:
 mov cx,1
@@6_:
 add edi,edx            
 add y1,cx
 mov ax,d
 cmp ax,0
 js @@4_                
 inc edi                
 add ax,aincr           
 jmp @@5_
@@4_:                   
 add ax,bincr           
@@5_:
 mov d,ax
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al           

 mov ax,y2
 cmp ax,y1
 jne @@6_               
 jmp @@Exit

@@Vertical:             
 mov ax,y1
 cmp ax,y2
 jbe @@7                
 mov bx,y2
 mov y1,bx
 mov y2,ax              
@@7:
 mov edi,_FrameBuffer   
 movzx eax,y1           
 movzx ecx,_LogicalScreenWidth          
 mul ecx                                
 mov dx,x1
 add eax,edx                            
 add edi,eax                    
 mov cx,y2
 sub cx,y1
 inc cx                         
 mov dx,_LogicalScreenWidth
@@8:
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al                   
 add edi,edx
 loop @@8

@@Exit:
end;

procedure transMaskedLine(x1,y1,x2,y2:word;Color:byte;Mask:word);
assembler;
var d,_dx,_dy,aincr,bincr,yincr:integer;
asm
 mov ax,x1
 cmp ax,x2
 je @@Vertical
 jbe @@1                
 mov bx,x2
 mov x1,bx
 mov x2,ax              
 mov ax,y1
 mov bx,y2
 mov y1,bx
 mov y2,ax              
@@1:
 mov cx,_LogicalScreenWidth
 mov ax,y1
 mov bx,y2
 cmp bx,ax
 ja @@2                 
 neg cx                 
@@2:
 mov yincr,cx           
 sub bx,ax              
 jns @@3
 neg bx
@@3:
 mov _dy,bx             
 mov ax,x2
 sub ax,x1
 mov _dx,ax             
 cmp ax,bx
 jb @@45                
 shl bx,1
 mov bincr,bx           
 sub bx,ax
 mov d,bx               
 mov ax,_dy
 sub ax,_dx
 jmp @@Continue
@@45:                   
 shl ax,1
 mov bincr,ax           
 sub ax,bx
 mov d,ax               
 mov ax,_dx
 sub ax,_dy
@@Continue:
 sal ax,1
 mov aincr,ax           

 movzx edx,_LogicalScreenWidth
 movzx eax,y1
 mul edx
 mov edi,_FrameBuffer
 add edi,eax
 mov dx,x1
 add edi,edx
 rol Mask,1
 jnc @@Skip1stPix
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al           
@@Skip1stPix:

 movsx edx,yincr
 mov ax,_dx
 cmp ax,_dy
 jb @@LessThen45

@@6:
 inc edi                
 inc x1
 mov ax,d
 cmp ax,0
 js @@4                 
 add edi,edx            
 add ax,aincr           
 jmp @@5
@@4:                    
 add ax,bincr           
@@5:
 mov d,ax
 rol Mask,1
 jnc @@SkipPix
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al           
@@SkipPix:

 mov ax,x2
 cmp ax,x1
 ja @@6                 
 jmp @@Exit

@@LessThen45:
 xor edx,0
 jns @@Positive
 mov cx,-1
 jmp @@6_
@@Positive:
 mov cx,1
@@6_:
 add edi,edx            
 add y1,cx
 mov ax,d
 cmp ax,0
 js @@4_                
 inc edi                
 add ax,aincr           
 jmp @@5_
@@4_:                   
 add ax,bincr           
@@5_:
 mov d,ax
 rol Mask,1
 jnc @@SkipPix_
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al           
@@SkipPix_:

 mov ax,y2
 cmp ax,y1
 jne @@6_               
 jmp @@Exit

@@Vertical:             
 mov ax,y1
 cmp ax,y2
 jbe @@7                
 mov bx,y2
 mov y1,bx
 mov y2,ax              
@@7:
 mov edi,_FrameBuffer   
 movzx eax,y1           
 movzx ecx,_LogicalScreenWidth           
 mul ecx                                 
 mov dx,x1
 add eax,edx                    
 add edi,eax                    
 mov cx,y2
 sub cx,y1
 inc cx                         
 mov dx,_LogicalScreenWidth
@@8:
 rol Mask,1
 jnc @@SkipPix__
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al                   
@@SkipPix__:
 add edi,edx
 loop @@8

@@Exit:
end;

procedure transHLine(x1,x2,y:word;Color:byte);assembler;
asm
 cld
 mov edi,_FrameBuffer           
 movzx eax,y
 movzx ecx,_LogicalScreenWidth  
 mul ecx                        
 mov dx,x1
 add eax,edx                    
 add edi,eax                    
 mov cx,x2
 sub cx,dx
 inc cx                         
@@1:
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 stosb
 loop @@1
end;

procedure transVLine(x,y1,y2:word;Color:byte);assembler;
asm
 mov edi,_FrameBuffer           
 movzx eax,y1                   
 movzx ecx,_LogicalScreenWidth  
 mul ecx                        
 mov dx,x
 add eax,edx                    
 add edi,eax                    
 mov cx,y2
 sub cx,y1
 inc cx                         
 mov dx,_LogicalScreenWidth
@@1:
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 mov [edi],al                   
 add edi,edx
 loop @@1
end;

procedure transRectangle(x1,y1,x2,y2:word;Color:byte);
begin
 transHLine(x1,x2,y1,Color);
 transHLine(x1,x2,y2,Color);
 transVLine(x1,y1,y2,Color);
 transVLine(x2,y1,y2,Color);
end;

procedure transBar(x1,y1,x2,y2:word;Color:byte);assembler;
var PixelsPerLine:word;
asm
 mov edi,_FrameBuffer
 movzx eax,y1
 movzx ecx,_LogicalScreenWidth
 mul ecx
 add edi,eax
 mov cx,x1
 add edi,ecx
 mov cx,x2
 sub cx,x1
 inc cx                         
 mov PixelsPerLine,cx
 mov dx,_LogicalScreenWidth
 sub dx,cx
@@0:
 mov cx,PixelsPerLine
@@1:
 movzx ebx,Color
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 stosb
 loop @@1
 add edi,edx
 inc y1
 mov cx,y2
 cmp cx,y1                      
 jns @@0
end;

procedure transEllipse(xc,yc,rx,ry:word;Color:byte);
var x,y:word;
    a,b,a2,b2:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  transVLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  transHLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx*rx;
 a2:=a shl 1; b:=ry*ry; b2:=b shl 1;
 d:=b-a*ry+(a shr 2); dx:=0; dy:=a2*ry;
 while dx<dy do begin
  transPutPixel(xc+x,yc+y,Color);
  transPutPixel(xc-x,yc+y,Color);
  transPutPixel(xc+x,yc-y,Color);
  transPutPixel(xc-x,yc-y,Color);
  if d>0 then begin y:=y-1; dy:=dy-a2; d:=d-dy; end;
  x:=x+1; dx:=dx+b2; d:=d+b+dx;
 end;
 d:=d+((3*(a-b)div 2-(dx+dy))div 2);
 while y>0 do begin
  transPutPixel(xc+x,yc+y,Color);
  transPutPixel(xc-x,yc+y,Color);
  transPutPixel(xc+x,yc-y,Color);
  transPutPixel(xc-x,yc-y,Color);
  if d<0 then begin x:=x+1; dx:=dx+b2; d:=d+dx; end;
  y:=y-1; dy:=dy-a2; d:=d+a-dy;
 end;
 transPutPixel(xc+x,yc,Color);
 transPutPixel(xc-x,yc,Color);
end;

procedure transFilledEllipse(xc,yc,rx,ry:word;Color:byte);
var x,y:word;
    a,b,a2,b2:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  transVLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  transHLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx*rx; b:=ry*ry; 
 a2:=a shl 1; b2:=b shl 1;
 d:=b-a*ry+(a shr 2); dx:=0; dy:=a2*ry;
 transVLine(xc,yc-y,yc+y,Color);
 while dx<dy do begin
  if d>0 then begin Dec(y); Dec(dy,a2); Dec(d,dy); end;
  Inc(x); Inc(dx,b2); Inc(d,b+dx);
  transVLine(xc-x,yc-y,yc+y,Color);
  transVLine(xc+x,yc-y,yc+y,Color);
 end;
 d:=d+((3*(a-b)div 2-(dx+dy))div 2);
 while y>0 do begin
  if d<0 then begin
   Inc(x); Inc(dx,b2); Inc(d,b+dx);
   transVLine(xc-x,yc-y,yc+y,Color);
   transVLine(xc+x,yc-y,yc+y,Color);
  end;
  Dec(y); Dec(dy,a2); Inc(d,a-dy);
 end;
end;

procedure transTriangle(x1,y1,x2,y2,x3,y3:word;Color:byte);
begin
 transLine(x1,y1,x2,y2,Color);
 transLine(x2,y2,x3,y3,Color);
 transLine(x3,y3,x1,y1,Color);
end;

procedure transFillTriangle(x1,y1,x2,y2,x3,y3:integer;Color:byte);
assembler;
var ymin,xmin,ymid,xmid,ymax,xmax,xmid2,i:longint;
asm
 mov ax,y1
 cmp ax,y2
 jle @@1
 xchg ax,y2
 mov y1,ax
 mov ax,x1
 xchg ax,x2
 mov x1,ax
@@1:
 mov ax,y1
 cmp ax,y3
 jle @@2
 xchg ax,y3
 mov y1,ax
 mov ax,x1
 xchg ax,x3
 mov x1,ax
@@2:
 mov ax,y2
 cmp ax,y3
 jle @@4
 xchg ax,y3
 mov y2,ax
 mov ax,x2
 xchg ax,x3
 mov x2,ax
@@4:
 movzx eax,y1
 mov ymin,eax
 movzx eax,x1
 mov xmin,eax
 movzx eax,y2
 mov ymid,eax
 movzx eax,x2
 mov xmid,eax
 movzx eax,y3
 mov ymax,eax
 movzx eax,x3
 mov xmax,eax
 sub eax,xmin
 mov ebx,ymid
 sub ebx,ymin
 imul ebx
 mov ebx,ymax
 sub ebx,ymin
 jz @@exit
 idiv ebx
 add eax,xmin
 mov xmid2,eax
 cmp eax,xmid
 jge @@5
 xchg eax,xmid
 mov xmid2,eax
@@5:
 mov eax,ymin
 cmp eax,ymid
 je @@6
 mov eax,ymin
 mov i,eax
@@loop1:
 mov eax,xmid
 sub eax,xmin
 mov ebx,i
 sub ebx,ymin
 imul ebx
 mov ebx,ymid
 sub ebx,ymin
 idiv ebx
 add eax,xmin
 push eax
 mov eax,xmid2
 sub eax,xmin
 mov ebx,i
 sub ebx,ymin
 imul ebx
 mov ebx,ymid
 sub ebx,ymin
 idiv ebx
 add eax,xmin
 push eax
 push i
 push dword ptr color
 call transHLine
 inc i
 mov eax,i
 sub eax,ymid
 cmp ax,2
 ja @@loop1
@@6:
 mov eax,ymid
 cmp eax,ymax
 je @@exit
 mov eax,ymid
 mov i,eax
@@loop2:
 mov eax,xmax
 sub eax,xmid
 mov ebx,i
 sub ebx,ymid
 imul ebx
 mov ebx,ymax
 sub ebx,ymid
 idiv ebx
 add eax,xmid
 push eax
 mov eax,xmax
 sub eax,xmid2
 mov ebx,i
 sub ebx,ymid
 imul ebx
 mov ebx,ymax
 sub ebx,ymid
 idiv ebx
 add eax,xmid2
 push eax
 push i
 push dword ptr Color
 call transHLine
 inc i
 mov eax,i
 sub eax,ymax
 cmp ax,0
 ja @@loop2
@@exit:
end;

procedure transFillPoly(var VertexArray;NumOfVertexes,Color:byte);
type TVertex=record
      x,y:integer;
     end;
     TVertexArray=array[1..256] of TVertex;
begin
 if NumOfVertexes=1 then begin
  transPutPixel(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,color);
  Exit;
 end;
 if NumOfVertexes=2 then begin
  transLine(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
       TVertexArray(VertexArray)[2].x,TVertexArray(VertexArray)[2].y,
            color);
  Exit;
 end;
 repeat
  transFillTriangle(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
          TVertexArray(VertexArray)[NumOfVertexes-1].x,TVertexArray(VertexArray)[NumOfVertexes-1].y,
          TVertexArray(VertexArray)[NumOfVertexes].x,TVertexArray(VertexArray)[NumOfVertexes].y,
                    Color);
  Dec(NumOfVertexes);
 until NumOfVertexes=2;
end;



procedure _transPutPixel(x,y:integer;Color:byte);
begin
 if (x>=_MinClipX) and
    (x<=_MaxClipX) and
    (y>=_MinClipY) and
    (y<=_MaxClipY) then transPutPixel(x,y,Color);
end;

procedure _transPutTransparentPixel(x,y:integer;Color:byte);
begin
 if (x>=_MinClipX) and
    (x<=_MaxClipX) and
    (y>=_MinClipY) and
    (y<=_MaxClipY) then transPutTransparentPixel(x,y,Color);
end;

procedure xchange(var a,b:integer);
var c:integer;
begin
 c:=a;a:=b;b:=c;
end;

function OutCode(x,y:integer):integer;
var code:integer;
begin
 code:=0;
 if x<_MinClipX then code:=(code or 1);
 if y<_MinClipY then code:=(code or 2);
 if x>_MaxClipX then code:=(code or 4);
 if y>_MaxClipY then code:=(code or 8);
 OutCode:=code;
end;

function ClipLine(var x1,y1,x2,y2:integer):boolean;

(* This is Sutherland - Cohen clipping algorithm *)

var code1,code2:integer;
    inside,outside:boolean;
begin
 code1:=OutCode(x1,y1);
 code2:=OutCode(x2,y2);
 inside:=(code1 or code2)=0;
 outside:=(code1 and code2)<>0;
 while (not inside) and (not outside) do begin
  if code1=0 then begin
   xchange(x1,x2);
   xchange(y1,y2);
   xchange(code1,code2);
  end;
  if (code1 and 1)<>0 then begin
   y1:=y1+longint(y2-y1)*(_MinClipX-x1) div (x2-x1);
   x1:=_MinClipX;
  end else
  if (code1 and 2)<>0 then begin
   x1:=x1+longint(x2-x1)*(_MinClipY-y1) div (y2-y1);
   y1:=_MinClipY;
  end else
  if (code1 and 4)<>0 then begin
   y1:=y1+longint(y2-y1)*(_MaxClipX-x1) div (x2-x1);
   x1:=_MaxClipX;
  end else
  if (code1 and 8)<>0 then begin
   x1:=x1+longint(x2-x1)*(_maxClipY-y1) div (y2-y1);
   y1:=_MaxClipY;
  end;
  code1:=OutCode(x1,y1);
  code2:=OutCode(x2,y2);
  inside:=(code1 or code2)=0;
  outside:=(code1 and code2)<>0;
 end;
 ClipLine:=(not outside);
end;

procedure _transLine(x1,y1,x2,y2:integer;Color:byte);
begin
 if ClipLine(x1,y1,x2,y2) then transLine(x1,y1,x2,y2,Color);
end;

procedure _transMaskedLine(x1,y1,x2,y2:integer;Color:byte;Mask:word);
var MaskShift:word;
begin
 if x1<_MinClipX then MaskShift:=(_MinClipX-x1) mod 16 else MaskShift:=0;
 if ClipLine(x1,y1,x2,y2) then transMaskedLine(x1,y1,x2,y2,Color,Mask shl MaskShift);
end;

procedure _transHLine(x1,x2,y:integer;Color:byte);
begin
 if (y<_MinClipY) or
    (y>_MaxClipY) or
    (x1>_MaxClipX) or
    (x2<_MinClipX) then Exit;
 if x1<_MinClipX then x1:=_MinClipX;
 if x2>_MaxClipX then x2:=_MaxClipX;
 transHLine(x1,x2,y,Color);
end;

procedure _transVLine(x,y1,y2:integer;Color:byte);
begin
 if (x<_MinClipX) or
    (x>_MaxClipX) or
    (y1>_MaxClipY) or
    (y2<_MinClipY) then Exit;
 if y1<_MinClipY then y1:=_MinClipY;
 if y2>_MaxClipY then y2:=_MaxClipY;
 transVLine(x,y1,y2,Color);
end;

procedure _transRectangle(x1,y1,x2,y2:integer;Color:byte);
begin
 _transHLine(x1,x2,y1,Color);
 _transVLine(x1,y1,y2,Color);
 _transVLine(x2,y1,y2,Color);
 _transHLine(x1,x2,y2,Color);
end;

procedure _transBar(x1,y1,x2,y2:integer;Color:byte);
begin
 if (x1>_MaxClipX) or
    (y1>_MaxClipY) or
    (x2<_MinClipX) or
    (y2<_MinClipY) then Exit;
 if x1<_MinClipX then x1:=_MinClipX;
 if y1<_MinClipY then y1:=_MinClipY;
 if x2>_MaxClipX then x2:=_MaxClipX;
 if y2>_MaxClipY then Y2:=_MaxClipY;
 transBar(x1,y1,x2,y2,Color);
end;

procedure _transEllipse(xc,yc:integer;rx,ry:word;Color:byte);
var x,y:word;
    a,b,a2,b2:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  _transVLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  _transHLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx*rx;
 a2:=a shl 1; b:=ry*ry; b2:=b shl 1;
 d:=b-a*ry+(a shr 2); dx:=0; dy:=a2*ry;
 while dx<dy do begin
  _transPutPixel(xc+x,yc+y,Color);
  _transPutPixel(xc-x,yc+y,Color);
  _transPutPixel(xc+x,yc-y,Color);
  _transPutPixel(xc-x,yc-y,Color);
  if d>0 then begin y:=y-1; dy:=dy-a2; d:=d-dy; end;
  x:=x+1; dx:=dx+b2; d:=d+b+dx;
 end;
 d:=d+((3*(a-b)div 2-(dx+dy))div 2);
 while y>0 do begin
  _transPutPixel(xc+x,yc+y,Color);
  _transPutPixel(xc-x,yc+y,Color);
  _transPutPixel(xc+x,yc-y,Color);
  _transPutPixel(xc-x,yc-y,Color);
  if d<0 then begin x:=x+1; dx:=dx+b2; d:=d+dx; end;
  y:=y-1; dy:=dy-a2; d:=d+a-dy;
 end;
 _transPutPixel(xc+x,yc,Color);
 _transPutPixel(xc-x,yc,Color);
end;

procedure _transFilledEllipse(xc,yc:integer;rx,ry:word;Color:byte);
var x,y:word;
    a,b,a2,b2:longint;
    d,dx,dy:longint;
begin
 if rx=0 then begin
  _transVLine(xc,yc-ry,yc+ry,Color); Exit;
 end;
 if ry=0 then begin
  _transHLine(xc-rx,xc+rx,yc,Color); Exit;
 end;
 x:=0; y:=ry; a:=rx*rx; b:=ry*ry; 
 a2:=a shl 1; b2:=b shl 1;
 d:=b-a*ry+(a shr 2); dx:=0; dy:=a2*ry;
 _transVLine(xc,yc-y,yc+y,Color);
 while dx<dy do begin
  if d>0 then begin Dec(y); Dec(dy,a2); Dec(d,dy); end;
  Inc(x); Inc(dx,b2); Inc(d,b+dx);
  _transVLine(xc-x,yc-y,yc+y,Color);
  _transVLine(xc+x,yc-y,yc+y,Color);
 end;
 d:=d+((3*(a-b)div 2-(dx+dy))div 2);
 while y>0 do begin
  if d<0 then begin
   Inc(x); Inc(dx,b2); Inc(d,b+dx);
   _transVLine(xc-x,yc-y,yc+y,Color);
   _transVLine(xc+x,yc-y,yc+y,Color);
  end;
  Dec(y); Dec(dy,a2); Inc(d,a-dy);
 end;
end;

procedure _transTriangle(x1,y1,x2,y2,x3,y3:integer;Color:byte);
begin
 _transLine(x1,y1,x2,y2,Color);
 _transLine(x2,y2,x3,y3,Color);
 _transLine(x3,y3,x1,y1,Color);
end;

procedure _transFillTriangle(x1,y1,x2,y2,x3,y3:integer;Color:byte);
var ymin,xmin,ymid,xmid,ymax,xmax,xmid2,i:longint;
asm
 mov ax,y1
 cmp ax,y2
 jle @@1
 xchg ax,y2
 mov y1,ax
 mov ax,x1
 xchg ax,x2
 mov x1,ax
@@1:
 mov ax,y1
 cmp ax,y3
 jle @@2
 xchg ax,y3
 mov y1,ax
 mov ax,x1
 xchg ax,x3
 mov x1,ax
@@2:
 mov ax,y2
 cmp ax,y3
 jle @@4
 xchg ax,y3
 mov y2,ax
 mov ax,x2
 xchg ax,x3
 mov x2,ax
@@4:
 movzx eax,y1
 mov ymin,eax
 movzx eax,x1
 mov xmin,eax
 movzx eax,y2
 mov ymid,eax
 movzx eax,x2
 mov xmid,eax
 movzx eax,y3
 mov ymax,eax
 movzx eax,x3
 mov xmax,eax
 sub eax,xmin
 mov ebx,ymid
 sub ebx,ymin
 imul ebx
 mov ebx,ymax
 sub ebx,ymin
 jz @@exit
 idiv ebx
 add eax,xmin
 mov xmid2,eax
 cmp eax,xmid
 jge @@5
 xchg eax,xmid
 mov xmid2,eax
@@5:
 mov eax,ymin
 cmp eax,ymid
 je @@6
 mov eax,ymin
 mov i,eax
@@loop1:
 mov eax,xmid
 sub eax,xmin
 mov ebx,i
 sub ebx,ymin
 imul ebx
 mov ebx,ymid
 sub ebx,ymin
 idiv ebx
 add eax,xmin
 push eax
 mov eax,xmid2
 sub eax,xmin
 mov ebx,i
 sub ebx,ymin
 imul ebx
 mov ebx,ymid
 sub ebx,ymin
 idiv ebx
 add eax,xmin
 push eax
 push i
 push dword ptr color
 call _transHLine
 inc i
 mov eax,i
 sub eax,ymid
 cmp ax,2
 ja @@loop1
@@6:
 mov eax,ymid
 cmp eax,ymax
 je @@exit
 mov eax,ymid
 mov i,eax
@@loop2:
 mov eax,xmax
 sub eax,xmid
 mov ebx,i
 sub ebx,ymid
 imul ebx
 mov ebx,ymax
 sub ebx,ymid
 idiv ebx
 add eax,xmid
 push eax
 mov eax,xmax
 sub eax,xmid2
 mov ebx,i
 sub ebx,ymid
 imul ebx
 mov ebx,ymax
 sub ebx,ymid
 idiv ebx
 add eax,xmid2
 push eax
 push i
 push dword ptr Color
 call _transHLine
 inc i
 mov eax,i
 sub eax,ymax
 cmp ax,0
 ja @@loop2
@@exit:
end;

procedure _transFillPoly(var VertexArray;NumOfVertexes,Color:byte);
type TVertex=record
      x,y:integer;
     end;
     TVertexArray=array[1..256] of TVertex;
begin
 if NumOfVertexes=1 then begin
  _transPutPixel(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,color);
  Exit;
 end;
 if NumOfVertexes=2 then begin
  _transLine(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
        TVertexArray(VertexArray)[2].x,TVertexArray(VertexArray)[2].y,
             Color);
  Exit;
 end;
 repeat
  _transFillTriangle(TVertexArray(VertexArray)[1].x,TVertexArray(VertexArray)[1].y,
           TVertexArray(VertexArray)[NumOfVertexes-1].x,TVertexArray(VertexArray)[NumOfVertexes-1].y,
           TVertexArray(VertexArray)[NumOfVertexes].x,TVertexArray(VertexArray)[NumOfVertexes].y,
                     Color);
  Dec(NumOfVertexes);
 until NumOfVertexes=2;
end;



procedure transPutImage(x,y,Width,Height:word;Address:pointer);
assembler;
asm
 cmp Height,0
 je @@Exit
 mov edi,_FrameBuffer
 movzx eax,y
 movzx ecx,_LogicalScreenWidth
 mul ecx
 mov dx,x
 add eax,edx
 add edi,eax                    

 mov dx,_LogicalScreenWidth
 sub dx,Width
 mov esi,Address                

@@1:
 mov cx,Width
@@2:
 movzx ebx,byte ptr [esi]
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 stosb
 inc esi
 loop @@2
 add edi,edx                            
 dec Height                             
 jnz @@1
@@Exit:
end;

procedure transPutTransparentImage(x,y,Width,Height:word;Address:pointer);
assembler;
asm
 cmp Height,0
 je @@Exit
 mov edi,_FrameBuffer
 movzx eax,y
 movzx ecx,_LogicalScreenWidth
 mul ecx
 mov dx,x
 add eax,edx
 add edi,eax                    

 mov dx,_LogicalScreenWidth
 sub dx,Width
 mov ah,_TransparentColor
 mov esi,Address                

@@1:
 mov cx,Width
@@3:
 lodsb
 cmp al,ah
 je @@2
 movzx ebx,byte ptr [edi]
 shl ebx,8
 add ebx,_Table
 xlat
 mov [edi],al
@@2:
 inc edi
 loop @@3
 add edi,edx            
 dec Height             
 jnz @@1
@@Exit:
end;



procedure _transPutImage(x,y:integer;Width,Height:word;Address:pointer);
var TopCut,LeftCut,RightCut,OldWidth:dword;
asm
 cmp Height,0
 je @@Exit
 mov TopCut,0
 mov LeftCut,0
 mov RightCut,0
 movzx eax,Width
 mov OldWidth,eax

 mov ax,y
 cmp ax,_MaxClipY
 jg @@Exit
 cmp ax,_MinClipY
 jge @@1
                
 mov bx,_MinClipY
 mov y,bx
 sub bx,ax
 mov word ptr TopCut,bx
 sub Height,bx
@@1:
 mov ax,y
 add ax,Height
 dec ax
 cmp ax,_MinClipY
 jl @@Exit
 cmp ax,_MaxClipY
 jle @@2
                
 mov bx,_MaxClipY
 sub bx,y
 inc bx
 mov Height,bx
@@2:
 mov ax,x
 cmp ax,_MaxClipX
 jg @@Exit
 cmp ax,_MinClipX
 jge @@3
                
 mov bx,_MinClipX
 mov x,bx
 sub bx,ax
 mov word ptr LeftCut,bx
 sub Width,bx
@@3:
 mov ax,x
 add ax,Width
 dec ax
 cmp ax,_MinClipX
 jl @@Exit
 cmp ax,_MaxClipX
 jle @@4
                
 mov bx,_MaxClipX
 sub bx,x
 inc bx
 mov Width,bx
 sub ax,_MaxClipX
 mov word ptr RightCut,ax
@@4:

 mov edi,_FrameBuffer
 movzx eax,y
 movzx ecx,_LogicalScreenWidth
 mul ecx
 mov dx,x
 add eax,edx
 add edi,eax                    

 mov esi,Address
 mov eax,TopCut
 mul OldWidth
 add esi,eax
 add esi,LeftCut
 mov dx,_LogicalScreenWidth
 sub dx,Width

@@5:
 mov cx,Width
@@6:
 movzx ebx,byte ptr [esi]
 mov bh,[edi]
 add ebx,_Table
 mov al,[ebx]
 stosb
 inc esi
 loop @@6
 add edi,edx                            
 add esi,LeftCut
 add esi,RightCut
 dec Height                             
 jnz @@5

@@Exit:
end;

procedure _transPutTransparentImage(x,y:integer;Width,Height:word;Address:pointer);
var TopCut,LeftCut,RightCut,OldWidth:dword;
asm
 cmp Height,0
 je @@Exit
 mov TopCut,0
 mov LeftCut,0
 mov RightCut,0
 movzx eax,Width
 mov OldWidth,eax

 mov ax,y
 cmp ax,_MaxClipY
 jg @@Exit
 cmp ax,_MinClipY
 jge @@1
                
 mov bx,_MinClipY
 mov y,bx
 sub bx,ax
 mov word ptr TopCut,bx
 sub Height,bx
@@1:
 mov ax,y
 add ax,Height
 dec ax
 cmp ax,_MinClipY
 jl @@Exit
 cmp ax,_MaxClipY
 jle @@2
                
 mov bx,_MaxClipY
 sub bx,y
 inc bx
 mov Height,bx
@@2:
 mov ax,x
 cmp ax,_MaxClipX
 jg @@Exit
 cmp ax,_MinClipX
 jge @@3
                
 mov bx,_MinClipX
 mov x,bx
 sub bx,ax
 mov word ptr LeftCut,bx
 sub Width,bx
@@3:
 mov ax,x
 add ax,Width
 dec ax
 cmp ax,_MinClipX
 jl @@Exit
 cmp ax,_MaxClipX
 jle @@4
                
 mov bx,_MaxClipX
 sub bx,x
 inc bx
 mov Width,bx
 sub ax,_MaxClipX
 mov word ptr RightCut,ax
@@4:

 mov edi,_FrameBuffer
 movzx eax,y
 movzx ecx,_LogicalScreenWidth
 mul ecx
 mov dx,x
 add eax,edx
 add edi,eax                    

 mov esi,Address                
 mov eax,TopCut
 mul OldWidth
 add esi,eax
 add esi,LeftCut                
 mov dx,_LogicalScreenWidth
 sub dx,Width
 mov ah,_TransparentColor

@@5:
 mov cx,Width
@@7:
 lodsb
 cmp al,bh              
 je @@6                 
 movzx ebx,byte ptr [edi]
 shl ebx,8
 add ebx,_Table
 xlat
 mov [edi],al
@@6:
 inc edi
 loop @@7
 add edi,edx            
 add esi,LeftCut
 add esi,RightCut
 dec Height             
 jnz @@5

@@Exit:
end;

type TFont=array[0..255,0..15] of byte;

procedure transWriteText(x,y:word;S:string;Color:byte);
var i,j:byte;
begin
 if _Font=nil then Exit;
 for i:=1 to Length(S) do begin
  for j:=0 to 15 do
   transMaskedLine(x,y+j,x+7,y+j,Color,word(TFont(_Font^)[Ord(S[i]),j]) shl 8);
  Inc(x,8);
 end;
end;

procedure _transWriteText(x,y:word;S:string;Color:byte);
var i,j:byte;
begin
 if _Font=nil then Exit;
 for i:=1 to Length(S) do begin
  for j:=0 to 15 do
   _transMaskedLine(x,y+j,x+7,y+j,Color,word(TFont(_Font^)[Ord(S[i]),j]) shl 8);
  Inc(x,8);
 end;
end;


end.

